from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any
from typing import Union

from pydantic import Field
from pydantic import computed_field
from pydantic import field_serializer

from zabbix_cli.models import TableRenderable
from zabbix_cli.pyzabbix.types import HostGroup
from zabbix_cli.pyzabbix.types import Template
from zabbix_cli.pyzabbix.types import TemplateGroup

if TYPE_CHECKING:
    from zabbix_cli.models import ColsRowsType
    from zabbix_cli.models import RowsType


class ShowTemplateGroupResult(TableRenderable):
    """Result type for templategroup."""

    groupid: str = Field(..., json_schema_extra={"header": "Group ID"})
    name: str
    templates: list[Template] = []
    show_templates: bool = Field(True, exclude=True)

    @classmethod
    def from_result(
        cls, group: Union[HostGroup, TemplateGroup], show_templates: bool
    ) -> ShowTemplateGroupResult:
        return cls(
            groupid=group.groupid,
            name=group.name,
            templates=group.templates,
            show_templates=show_templates,
        )

    @computed_field
    @property
    def template_count(self) -> int:
        return len(self.templates)

    @field_serializer("templates")
    def templates_serializer(self, value: list[Template]) -> list[dict[str, Any]]:
        if self.show_templates:
            return [t.model_dump(mode="json") for t in value]
        return []

    def __rows__(self) -> RowsType:
        tpls = self.templates if self.show_templates else []
        return [
            [
                self.groupid,
                self.name,
                "\n".join(str(t.host) for t in sorted(tpls, key=lambda t: t.host)),
                str(self.template_count),
            ]
        ]


class ExtendTemplateGroupResult(TableRenderable):
    source: str
    destination: list[str]
    templates: list[str]

    @classmethod
    def from_result(
        cls,
        src_group: Union[HostGroup, TemplateGroup],
        dest_group: Union[list[HostGroup], list[TemplateGroup]],
        templates: list[Template],
    ) -> ExtendTemplateGroupResult:
        return cls(
            source=src_group.name,
            destination=[grp.name for grp in dest_group],
            templates=[t.host for t in templates],
        )


class MoveTemplatesResult(TableRenderable):
    """Result type for `move_templates` command."""

    source: str
    destination: str
    templates: list[str]

    @classmethod
    def from_result(
        cls,
        source: Union[HostGroup, TemplateGroup],
        destination: Union[HostGroup, TemplateGroup],
    ) -> MoveTemplatesResult:
        return cls(
            source=source.name,
            destination=destination.name,
            templates=[template.host for template in source.templates],
        )

    def __cols_rows__(self) -> ColsRowsType:
        """Only print the template names in the table.

        Source and destination are apparent from the surrounding context.
        """
        cols = ["Templates"]
        rows: RowsType = [["\n".join(self.templates)]]
        return cols, rows


class TemplateGroupPermissions(TableRenderable):
    """Result type for template group permissions."""

    groupid: str
    name: str
    permissions: list[str]

    def __cols_rows__(self) -> ColsRowsType:
        cols = ["GroupID", "Name", "Permissions"]
        rows: RowsType = [[self.groupid, self.name, "\n".join(self.permissions)]]
        return cols, rows



from __future__ import annotations

import os

import pytest
from zabbix_cli.output.prompts import HEADLESS_VARS_SET
from zabbix_cli.output.prompts import TRUE_ARGS
from zabbix_cli.output.prompts import is_headless


@pytest.mark.parametrize("envvar", HEADLESS_VARS_SET)
@pytest.mark.parametrize("value", TRUE_ARGS)
def test_is_headless_set_true(envvar: str, value: str):
    """Returns True when the envvar is set to a truthy value."""
    _do_test_is_headless(envvar, value, True)


@pytest.mark.parametrize("envvar", HEADLESS_VARS_SET)
@pytest.mark.parametrize("value", ["0", "false", "", None])
def test_is_headless_set_false(envvar: str, value: str):
    """Returns False when the envvar is set to a falsey value or unset"""
    _do_test_is_headless(envvar, value, False)


@pytest.mark.parametrize(
    "envvar, value, expected",
    [
        ("DEBIAN_FRONTEND", "noninteractive", True),
        ("DEBIAN_FRONTEND", "teletype", False),
        ("DEBIAN_FRONTEND", "readline", False),
        ("DEBIAN_FRONTEND", "dialog", False),
        ("DEBIAN_FRONTEND", "gtk", False),
        ("DEBIAN_FRONTEND", "text", False),
        ("DEBIAN_FRONTEND", "anything", False),
        ("DEBIAN_FRONTEND", None, False),
        ("DEBIAN_FRONTEND", "", False),
        ("DEBIAN_FRONTEND", "0", False),
        ("DEBIAN_FRONTEND", "false", False),
        ("DEBIAN_FRONTEND", "1", False),
        ("DEBIAN_FRONTEND", "true", False),
    ],
)
def test_is_headless_map(envvar: str, value: str, expected: bool) -> None:
    """Returns True when the envvar is set to a specific value."""
    _do_test_is_headless(envvar, value, expected)


def _do_test_is_headless(envvar: str, value: str | None, expected: bool):
    """Helper function for testing is_headless.

    Sets/clears envvar before testing, then clears cache and envvar after test.
    """
    _orig_environ = os.environ.copy()
    os.environ.clear()
    try:
        if value is None:
            os.environ.pop(envvar, None)
        else:
            os.environ[envvar] = value
        assert is_headless() == expected
    finally:
        # IMPORTANT: Remove envvar and clear cache after each test
        os.environ = _orig_environ  # type: ignore  # noqa: B003 # I _think_ this is fine?
        is_headless.cache_clear()



from __future__ import annotations

from pathlib import Path
from typing import Any
from typing import Optional
from typing import Union

import pytest

try:
    import tomllib
except ImportError:
    import tomli as tomllib

from inline_snapshot import snapshot
from pydantic import BaseModel
from pydantic import Field
from pydantic import model_validator
from typing_extensions import Self
from zabbix_cli.config.commands import CommandConfig
from zabbix_cli.config.commands import CreateHostOrTemplateGroup
from zabbix_cli.config.constants import OutputFormat
from zabbix_cli.config.constants import SecretMode
from zabbix_cli.config.model import APIConfig
from zabbix_cli.config.model import AppConfig
from zabbix_cli.config.model import AutoUpdateConfig
from zabbix_cli.config.model import Config
from zabbix_cli.config.model import PluginConfig
from zabbix_cli.config.model import PluginsConfig
from zabbix_cli.config.utils import DeprecatedField
from zabbix_cli.config.utils import fmt_deprecated_fields
from zabbix_cli.config.utils import get_deprecated_fields_set
from zabbix_cli.config.utils import replace_deprecated_fields
from zabbix_cli.exceptions import ConfigOptionNotFound
from zabbix_cli.exceptions import PluginConfigTypeError


def test_config_default() -> None:
    """Assert that the config can be instantiated with no arguments."""
    assert Config()


def test_sample_config() -> None:
    """Assert that the sample config can be instantiated."""
    assert Config.sample_config()


@pytest.mark.parametrize(
    "bespoke",
    [True, False],
)
def test_load_config_file_legacy(legacy_config_path: Path, bespoke: bool) -> None:
    if bespoke:
        conf = Config.from_conf_file(legacy_config_path)
    else:
        conf = Config.from_file(legacy_config_path)
    assert conf
    # Should be loaded from the file we specified
    assert conf.config_path == legacy_config_path
    # Should be marked as legacy
    assert conf.app.is_legacy is True
    # Should use legacy JSON format automatically
    assert conf.app.legacy_json_format is True


def remove_path_options(config_path: Path, tmp_path: Path) -> None:
    """Remove all path options from a TOML config file.

    Some config options require a directory or file to exist, which is not always
    possible or desirable in a test environment.
    """
    contents = config_path.read_text()
    new_contents = "\n".join(
        line for line in contents.splitlines() if "/path/to" not in line
    )
    config_path.write_text(new_contents)


def replace_paths(config_path: Path, tmp_path: Path) -> None:
    """Replace all /path/to paths in a file with temporary directories."""
    contents = config_path.read_text()
    new_contents = contents.replace("/path/to", str(tmp_path))
    tmp_path.mkdir(exist_ok=True)
    config_path.write_text(new_contents)


@pytest.mark.parametrize(
    "bespoke",
    [True, False],
)
@pytest.mark.parametrize(
    "with_paths",
    [True, False],
)
def test_load_config_file(
    config_path: Path, tmp_path: Path, bespoke: bool, with_paths: bool
) -> None:
    """Test loading a TOML configuration file."""
    # Test with and without custom file paths
    if with_paths:
        replace_paths(config_path, tmp_path)
    else:
        remove_path_options(config_path, tmp_path)

    # Use bespoke method for loading the given format
    if bespoke:
        conf = Config.from_toml_file(config_path)
    else:
        conf = Config.from_file(config_path)

    assert conf
    # Should be loaded from the file we specified
    assert conf.config_path == config_path
    assert conf.app.is_legacy is False
    assert conf.app.legacy_json_format is False


def test_plugins_config_get() -> None:
    """Test that we can get a plugin configuration."""
    config = PluginsConfig(
        root={
            "test1": PluginConfig(
                enabled=True,
                module="zabbix_cli.plugins.test1",
            ),
            "test2": PluginConfig(
                enabled=True,
                module="zabbix_cli.plugins.test2",
            ),
        }
    )
    assert config.get("test1")
    assert config.get("test2")
    assert not config.get("test3")


def test_plugin_config_get() -> None:
    config = PluginConfig(module="test")
    assert config.get("module") == "test"

    # With type validation
    assert config.get("module", type=str) == "test"

    # Missing option
    with pytest.raises(ConfigOptionNotFound):
        assert config.get("missing") is None

    # Default value
    assert config.get("missing", "default") == "default"

    # Default value with type validation
    assert config.get("missing", "default", type=str) == "default"

    # Extra values
    config = PluginConfig(
        module="test",
        extra1="foo",
        extra2=2,
        extra3=True,
        extra4=[1, 2, 3],
    )
    assert config.get("extra1") == "foo"
    assert config.get("extra2") == 2
    assert config.get("extra3") is True
    assert config.get("extra4") == [1, 2, 3]

    # With type validation
    assert config.get("extra1", type=str) == "foo"
    assert config.get("extra2", type=int) == 2
    assert config.get("extra3", type=bool) is True
    assert config.get("extra4", type=list) == [1, 2, 3]

    # Type validation with coercion
    assert config.get("extra3", type=int) == 1
    assert config.get("extra4", type=tuple) == (1, 2, 3)
    # Cannot coerce int to str by default in Pydantic
    with pytest.raises(PluginConfigTypeError):
        assert config.get("extra2", type=str) == "2"


def test_config_get_with_annotations() -> None:
    """Test PluginConfig.get with more complex annotations"""
    config = PluginConfig(
        module="test",
        extra1="foo",
        extra2=2,
        extra3=True,
        extra4=[1, 2, 3],
        extra5={"foo": [1, 2, 3]},
    )

    assert config.get("extra1", type=Optional[str]) == "foo"
    assert config.get("extra1", 123, type=Optional[str]) == "foo"
    assert config.get("extra1", type=Union[str, int])

    # Invalid default type
    with pytest.raises(PluginConfigTypeError):
        config.get("wrong", 123, type=Optional[str])
    assert config.get("wrong", "123", type=Optional[str]) == "123"

    # List type
    assert config.get("extra4", type=list) == [1, 2, 3]
    assert config.get("extra4", type=list[int]) == [1, 2, 3]
    assert config.get("extra4", type=list[int]) == [1, 2, 3]

    # Dict type
    assert config.get("extra5", type=dict) == {"foo": [1, 2, 3]}
    assert config.get("extra5", type=dict[str, list[int]]) == {"foo": [1, 2, 3]}
    assert config.get("extra5", type=dict[str, list[int]]) == {"foo": [1, 2, 3]}


def test_plugin_config_set() -> None:
    config = PluginConfig()

    # Existing model field
    config.set("module", "new")
    assert config.module == "new"

    # New attribute/field
    config.set("extra", "new")
    assert config.extra == "new"

    # Any type goes
    obj = object()
    config.set("object", obj)
    assert config.object is obj

    # Instantiated with extra values
    config = PluginConfig(
        module="test",
        extra1="extra1",
        extra2=2,
        extra3=True,
    )

    # Override existing values
    config.set("extra1", "new")
    config.set("extra2", 3)
    config.set("extra3", False)

    assert config.extra1 == "new"
    assert config.extra2 == 3
    assert config.extra3 is False

    assert config.model_dump() == snapshot(
        {
            "module": "test",
            "enabled": True,
            "optional": False,
            "extra1": "new",
            "extra2": 3,
            "extra3": False,
        }
    )


@pytest.mark.parametrize(
    "context, expect",
    [
        ({"secrets": SecretMode.HIDE}, SecretMode.HIDE),
        ({"secrets": SecretMode.MASK}, SecretMode.MASK),
        ({"secrets": SecretMode.PLAIN}, SecretMode.PLAIN),
        ({"secrets": "hide"}, SecretMode.HIDE),
        ({"secrets": "mask"}, SecretMode.MASK),
        ({"secrets": "plain"}, SecretMode.PLAIN),
        ({}, SecretMode.MASK),
        (None, SecretMode.MASK),
        (lambda: {}, SecretMode.MASK),  # type: ignore
        ({"secrets": True}, SecretMode.PLAIN),
        ({"secrets": False}, SecretMode.MASK),
    ],
)
def test_secret_mode_from_context(context: Any, expect: SecretMode) -> None:
    assert SecretMode.from_context(context) == expect


def test_secret_mode_default() -> None:
    assert SecretMode._DEFAULT is SecretMode.MASK  # type: ignore


def _get_config_with_secrets() -> Config:
    return Config(
        api=APIConfig(
            username="Admin",
            password="password",
            auth_token="token123",
        )
    )


def test_config_dump_to_file_masked(tmp_path: Path) -> None:
    conf_file = tmp_path / "zabbix-cli.toml"
    conf = _get_config_with_secrets()
    conf.dump_to_file(conf_file, secrets=SecretMode.MASK)

    toml_str = conf_file.read_text()
    config_dict = tomllib.loads(toml_str)
    assert config_dict["api"]["password"] == snapshot("**********")
    assert config_dict["api"]["auth_token"] == snapshot("**********")


def test_deprecated_field_warnings(caplog: pytest.LogCaptureFixture) -> None:
    """Test that deprecated fields are logged."""
    Config(
        app=AppConfig(
            output_format=OutputFormat.JSON,
            use_colors=True,
            use_paging=True,
            system_id="System-User",
            config_auto_update=AutoUpdateConfig(enabled=False),
        )
    )
    assert caplog.record_tuples == snapshot(
        [
            (
                "zabbix_cli",
                30,
                "Config option [configopt]output_format[/] is deprecated. Replaced by: [configopt]app.output.format[/].",
            ),
            (
                "zabbix_cli",
                30,
                "Config option [configopt]system_id[/] is deprecated. Replaced by: [configopt]api.username[/].",
            ),
            (
                "zabbix_cli",
                30,
                "Config option [configopt]use_colors[/] is deprecated. Replaced by: [configopt]app.output.color[/].",
            ),
            (
                "zabbix_cli",
                30,
                "Config option [configopt]use_paging[/] is deprecated. Replaced by: [configopt]app.output.paging[/].",
            ),
            (
                "zabbix_cli",
                30,
                """\
Your configuration file contains deprecated options.
  To update your config file with the new options, run:
  [command]zabbix-cli update_config[/]
  For more information, see the documentation.\
""",
            ),
        ]
    )


def test_get_deprecated_fields_set() -> None:
    config = Config(
        app=AppConfig(
            output_format=OutputFormat.JSON,
            use_colors=True,
            use_paging=True,
            system_id="System-User",
        )
    )
    # Sort because order of `BaseModel.model_fields_set` is not guaranteed
    assert sorted(get_deprecated_fields_set(config)) == snapshot(
        [
            DeprecatedField(
                field_name="app.output_format",
                value=OutputFormat.JSON,
                replacement=["app.output.format"],
            ),
            DeprecatedField(
                field_name="app.system_id",
                value="System-User",
                replacement=["api.username"],
            ),
            DeprecatedField(
                field_name="app.use_colors",
                value=True,
                replacement=["app.output.color"],
            ),
            DeprecatedField(
                field_name="app.use_paging",
                value=True,
                replacement=["app.output.paging"],
            ),
        ]
    )


def test_get_deprecated_fields_deep_nesting() -> None:
    """Test that deprecated fields are found in nested models."""

    class Baz(BaseModel):
        qux: str = ""

    class Bar(BaseModel):
        baz: Baz = Field(default_factory=Baz)
        qux_deprecated: str = Field(
            default="",
            deprecated=True,
            json_schema_extra={"replacement": "bar.baz.qux"},
        )

    class Foo(BaseModel):
        bar: Bar = Field(default_factory=Bar)

        # Copied from zabbix_cli.config.model.Config
        # TODO: to test this better, we could declare a new BaseModel
        # subclass intended for top-level models which automatically implement
        # this validator. That way we can be sure the validator behavior
        # in the test is consistent with the behavior in the actual code.
        @model_validator(mode="after")
        def _set_deprecated_fields_in_new_location(self) -> Self:
            """Set deprecated fields in their new location."""
            replace_deprecated_fields(self)
            return self

    foo = Foo(bar=Bar(qux_deprecated="test123"))

    # Deprecated field should be found
    assert sorted(get_deprecated_fields_set(foo)) == snapshot(
        [
            DeprecatedField(
                field_name="bar.qux_deprecated",
                value="test123",
                replacement=["bar.baz.qux"],
            )
        ]
    )

    # Deprecated field should be automatically assigned to the new field
    assert foo.model_dump() == snapshot(
        {"bar": {"baz": {"qux": "test123"}, "qux_deprecated": "test123"}}
    )


def test_deprecated_fields_updated() -> None:
    """Test that deprecated fields are assigned to the new fields if specified."""
    conf = Config(
        app=AppConfig(
            # Set fields to non-default values for comparison
            output_format=OutputFormat.JSON,
            use_colors=False,
            use_paging=True,
            system_id="System-User",
        )
    )
    assert sorted(conf.app.model_fields_set) == snapshot(
        ["output_format", "system_id", "use_colors", "use_paging"]
    )
    assert sorted(conf.app.output.model_fields_set) == snapshot(
        ["color", "format", "paging"]
    )
    assert conf.app.output.format == OutputFormat.JSON
    assert conf.app.output.color is False
    assert conf.app.output.paging is True
    assert conf.api.username == "System-User"


def get_deprecated_fields(model: Union[type[BaseModel], BaseModel]) -> list[str]:
    """Get a set of names of deprecated fields in a model and its submodels."""
    fields: list[str] = []
    for field_name, field in model.model_fields.items():
        if field.deprecated:
            fields.append(field_name)
        if not field.annotation:
            continue
        try:
            if issubclass(field.annotation, BaseModel):
                submodel_fields = get_deprecated_fields(field.annotation)
                fields.extend(
                    f"{field_name}.{subfield}" for subfield in submodel_fields
                )
        except TypeError:
            pass
    return fields


def test_get_deprecated_fields() -> None:
    """Ensure we are aware of all deprecated fields in the Config model"""
    assert get_deprecated_fields(Config) == snapshot(
        [
            "app.auth_token_file",
            "app.output_format",
            "app.use_colors",
            "app.use_paging",
            "app.system_id",
            "app.default_hostgroups",
            "app.default_admin_usergroups",
            "app.default_create_user_usergroups",
            "app.default_notification_users_usergroups",
            "app.export_directory",
            "app.export_format",
            "app.export_timestamps",
        ]
    )


def test_load_deprecated_config(tmp_path: Path) -> None:
    conf = tmp_path / "zabbix-cli.toml"
    conf.write_text(
        f"""
[api]
url = "http://localhost:8080"

# No username option specified (assigned from app.system_id)

password = ""
auth_token = ""
verify_ssl = true

[app]
use_session_file = true
auth_token_file = "{tmp_path}/.zabbix-cli_auth_token"
auth_file = "{tmp_path}/.zabbix-cli_auth"
history = true
history_file = "{tmp_path}/history"
bulk_mode = "strict"

# Deprecated options (moved)
default_hostgroups = ["All-hosts"]
default_admin_usergroups = ["All-admin-users"]
default_create_user_usergroups = ["All-users"]
default_notification_users_usergroups = ["All-notification-users"]
export_directory = "{tmp_path}/exports"
export_format = "json"
export_timestamps = false
use_colors = false
use_paging = true
output_format = "json"
system_id = "System-User"

# Deprecated options (slated for removal)
allow_insecure_auth_file = true
legacy_json_format = false

# No [app.output] section specified (assigned from deprecated app options)

[logging]
enabled = true
log_level = "DEBUG"
log_file = "{tmp_path}/zabbix-cli.log"

[plugins]
"""
    )
    # Check that we can actually load the config
    config = Config.from_file(conf)

    # Check that the deprecated fields are assigned to the new fields
    assert config.app.commands.create_host.hostgroups == ["All-hosts"]
    assert config.app.commands.create_hostgroup.rw_groups == ["All-admin-users"]
    assert config.app.commands.create_hostgroup.ro_groups == ["All-users"]
    assert config.app.commands.create_notification_user.usergroups == [
        "All-notification-users"
    ]
    assert config.app.commands.create_user.usergroups == ["All-users"]
    assert config.app.commands.export.directory == tmp_path / "exports"
    assert config.app.commands.export.format == OutputFormat.JSON
    assert config.app.commands.export.timestamps is False

    assert config.app.output.color is False
    assert config.app.output.paging is True
    assert config.app.output.format == OutputFormat.JSON
    assert config.api.username == "System-User"  # assigned from app.system_id


def test_load_deprecated_config_with_new_and_old_options(tmp_path: Path) -> None:
    """Test loading a config file where both new and deprecated options are present.

    The deprecated options should _not_ be assigned to the new options, as the new options
    are already set
    """
    conf = tmp_path / "zabbix-cli.toml"
    conf.write_text(
        f"""
##### Deprecated options (should be ignored) #####

[app]
default_hostgroups = ["All-hosts-deprecated"]
default_admin_usergroups = ["All-admin-users-deprecated"]
default_create_user_usergroups = ["All-users-deprecated"]
default_notification_users_usergroups = ["All-notification-users-deprecated"]
export_directory = "{tmp_path}/exports_deprecated"
export_format = "yaml"
export_timestamps = true
use_colors = false
use_paging = true
output_format = "json"
system_id = "System-User"

##### New options (should be used) #####

[api]
username = "Admin"

[app.commands.create_user]
usergroups = ["All-users"]

[app.commands.create_notification_user]
usergroups = ["All-notification-users"]

[app.commands.create_host]
hostgroups = ["All-hosts"]

[app.commands.create_hostgroup]
ro_groups = ["All-users"]
rw_groups = ["All-admin-users"]

[app.commands.export]
directory = "{tmp_path}/exports"
format = "json"
timestamps = false

[app.output]
color = true
format = "table"
paging = false
"""
    )
    config = Config.from_file(conf)

    # New fields should NOT be overwritten by deprecated fields
    assert config.app.commands.create_host.hostgroups == ["All-hosts"]
    assert config.app.commands.create_hostgroup.rw_groups == ["All-admin-users"]
    assert config.app.commands.create_hostgroup.ro_groups == ["All-users"]
    assert config.app.commands.create_user.usergroups == ["All-users"]
    assert config.app.commands.create_notification_user.usergroups == [
        "All-notification-users"
    ]
    assert config.app.commands.create_hostgroup.ro_groups == ["All-users"]
    assert config.app.commands.create_hostgroup.rw_groups == ["All-admin-users"]
    assert config.app.commands.export.directory == tmp_path / "exports"
    assert config.app.commands.export.format == OutputFormat.JSON
    assert config.app.commands.export.timestamps is False
    assert config.api.username == "Admin"
    assert config.app.output.color is True
    assert config.app.output.paging is False
    assert config.app.output.format == OutputFormat.TABLE


def test_load_deprecated_config_legacy(legacy_config_path: Path) -> None:
    """Test loading a legacy .conf config file with deprecated options."""
    config_str = legacy_config_path.read_text()
    assert "system_id=Test" in config_str
    assert "use_colors=ON" in config_str
    assert "use_paging=OFF" in config_str

    # Manipulate config to set default boolean values to opposite
    config_str = config_str.replace("use_colors=ON", "use_colors=OFF")
    config_str = config_str.replace("use_paging=OFF", "use_paging=ON")
    legacy_config_path.write_text(config_str)

    config = Config.from_conf_file(legacy_config_path)

    # Check that the deprecated fields are assigned to the new fields
    assert config.api.username == "Test"
    assert config.app.output.color is False
    assert config.app.output.paging is True

    # Check that the assigned fields are counted as set
    assert "username" in config.api.model_fields_set
    assert "color" in config.app.output.model_fields_set
    assert "paging" in config.app.output.model_fields_set
    assert "format" not in config.app.output.model_fields_set

    # Check that the deprecated fields are also set
    assert "system_id" in config.app.model_fields_set
    assert "use_colors" in config.app.model_fields_set
    assert "use_paging" in config.app.model_fields_set
    assert "output_format" not in config.app.model_fields_set


def test_shared_command_config(tmp_path: Path) -> None:
    """Test that the shared command config is loaded correctly."""

    # Test via kwargs
    config = Config(
        app=AppConfig(
            commands=CommandConfig(
                create_group=CreateHostOrTemplateGroup(
                    ro_groups=["All-users", "Guests"],
                    rw_groups=["All-admin-users", "All-techs"],
                )
            )
        )
    )
    assert config.app.commands.create_hostgroup.ro_groups == snapshot(
        ["All-users", "Guests"]
    )
    assert config.app.commands.create_hostgroup.rw_groups == snapshot(
        ["All-admin-users", "All-techs"]
    )
    assert config.app.commands.create_templategroup.ro_groups == snapshot(
        ["All-users", "Guests"]
    )
    assert config.app.commands.create_templategroup.rw_groups == snapshot(
        ["All-admin-users", "All-techs"]
    )

    # Test via config file
    conf_str = """
    [app.commands.create_group]
    ro_groups = ["All-users", "Guests"]
    rw_groups = ["All-admin-users", "All-techs"]
    """
    conf_path = tmp_path / "zabbix-cli.toml"
    conf_path.write_text(conf_str)
    config_shared = Config.from_toml_file(conf_path)

    assert config_shared.app.commands.create_hostgroup.ro_groups == snapshot(
        ["All-users", "Guests"]
    )
    assert config_shared.app.commands.create_hostgroup.rw_groups == snapshot(
        ["All-admin-users", "All-techs"]
    )
    assert config_shared.app.commands.create_templategroup.ro_groups == snapshot(
        ["All-users", "Guests"]
    )
    assert config_shared.app.commands.create_templategroup.rw_groups == snapshot(
        ["All-admin-users", "All-techs"]
    )

    # Test via config file (create_hostgroup is set: only create_templategroup uses shared)
    conf_str = """
    [app.commands.create_hostgroup]
    ro_groups = ["All-users"]
    rw_groups = ["All-admin-users"]

    [app.commands.create_group]
    ro_groups = ["All-users", "Guests"]
    rw_groups = ["All-admin-users", "All-techs"]
    """
    conf_path = tmp_path / "zabbix-cli_hg_set.toml"
    conf_path.write_text(conf_str)
    conf_hg_set = Config.from_toml_file(conf_path)

    assert conf_hg_set.app.commands.create_hostgroup.ro_groups == snapshot(
        ["All-users"]
    )
    assert conf_hg_set.app.commands.create_hostgroup.rw_groups == snapshot(
        ["All-admin-users"]
    )
    assert conf_hg_set.app.commands.create_templategroup.ro_groups == snapshot(
        ["All-users", "Guests"]
    )
    assert conf_hg_set.app.commands.create_templategroup.rw_groups == snapshot(
        ["All-admin-users", "All-techs"]
    )

    # Test via config file (create_templategroup is set: only create_hostgroup uses shared)
    conf_str = """
    [app.commands.create_templategroup]
    ro_groups = ["All-users"]
    rw_groups = ["All-admin-users"]

    [app.commands.create_group]
    ro_groups = ["All-users", "Guests"]
    rw_groups = ["All-admin-users", "All-techs"]
    """
    conf_path = tmp_path / "zabbix-cli_tg_set.toml"
    conf_path.write_text(conf_str)
    conf_tg_set = Config.from_toml_file(conf_path)

    assert conf_tg_set.app.commands.create_hostgroup.ro_groups == snapshot(
        ["All-users", "Guests"]
    )
    assert conf_tg_set.app.commands.create_hostgroup.rw_groups == snapshot(
        ["All-admin-users", "All-techs"]
    )
    assert conf_tg_set.app.commands.create_templategroup.ro_groups == snapshot(
        ["All-users"]
    )
    assert conf_tg_set.app.commands.create_templategroup.rw_groups == snapshot(
        ["All-admin-users"]
    )


def test_config_command_alias_priority(
    tmp_path: Path, caplog: pytest.LogCaptureFixture
) -> None:
    """Test loading a config file with duplicate sections with different aliases"""
    conf_import_first = """\
[app.commands.import]
format = "csv"

[app.commands.export]
format = "json"
"""
    conf_export_first = """\
[app.commands.export]
format = "json"

[app.commands.import]
format = "csv"
"""

    # Assert that the `import` alias is never used when `export` is defined
    for conf_str in [conf_import_first, conf_export_first]:
        conf_path = tmp_path / "zabbix-cli.toml"
        conf_path.write_text(conf_str)

        with caplog.at_level("WARNING", logger="zabbix_cli.config.commands"):
            conf = Config.from_toml_file(conf_path)

            # Uses the first alias found in the file regardless
            assert conf.app.commands.export.format == OutputFormat.JSON

            # The log record tuple should be the same for both cases
            assert caplog.record_tuples == snapshot(
                [
                    (
                        "zabbix_cli.config.commands",
                        30,
                        "Multiple export/import configuration sections found ([app.commands.export], [app.commands.import]). Using section: [app.commands.export]",
                    )
                ]
            )
            caplog.clear()


def test_fmt_deprecated_fields() -> None:
    deprecated_fields = [
        # A subset of deprecated fields for testing
        DeprecatedField(
            field_name="app.output_format",
            value=OutputFormat.JSON,
            replacement=["app.output.format"],
        ),
        DeprecatedField(
            field_name="app.system_id",
            value="System-User",
            replacement=["api.username"],
        ),
        DeprecatedField(
            field_name="app.use_colors",
            value=True,
            replacement=["app.output.color"],
        ),
        DeprecatedField(
            field_name="app.use_paging",
            value=True,
            replacement=["app.output.paging"],
        ),
        DeprecatedField(
            field_name="default_create_user_usergroups",
            value=[],
            replacement=[
                "app.commands.create_user.usergroups",
                "app.commands.create_hostgroup.ro_groups",
            ],
        ),
    ]
    formatted = fmt_deprecated_fields(deprecated_fields)
    assert formatted == snapshot(
        """\
  - [option]app.output_format[/] (replaced by: [option]app.output.format[/])
  - [option]app.system_id[/] (replaced by: [option]api.username[/])
  - [option]app.use_colors[/] (replaced by: [option]app.output.color[/])
  - [option]app.use_paging[/] (replaced by: [option]app.output.paging[/])
  - [option]default_create_user_usergroups[/] (replaced by: [option]app.commands.create_user.usergroups[/], [option]app.commands.create_hostgroup.ro_groups[/])\
"""
    )



# type: ignore
"""Patching of Typer to extend functionality and change styling.

Will probably break for some version of Typer at some point.
"""

from __future__ import annotations

import inspect
from collections.abc import Iterable
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import Union
from typing import cast
from uuid import UUID

import click
import typer
from typer.main import lenient_issubclass
from typer.models import ParameterInfo
import typer.rich_utils

from zabbix_cli._patches.common import get_patcher
from zabbix_cli.commands.common.args import CommandParam
from zabbix_cli.pyzabbix.enums import APIStrEnum

if TYPE_CHECKING:
    from rich.style import Style

patcher = get_patcher(f"Typer version: {typer.__version__}")


def patch_help_text_style() -> None:
    """Remove dimming of help text.

    https://github.com/tiangolo/typer/issues/437#issuecomment-1224149402
    """
    typer.rich_utils.STYLE_HELPTEXT = ""


def patch_help_text_spacing() -> None:
    """Adds a single blank line between short and long help text of a command when using `--help`.

    As of Typer 0.9.0, the short and long help text is printed without any
    blank lines between them. This is bad for readability (IMO).
    """
    from rich.console import group
    from rich.markdown import Markdown
    from rich.text import Text
    from typer.rich_utils import DEPRECATED_STRING
    from typer.rich_utils import MARKUP_MODE_MARKDOWN
    from typer.rich_utils import MARKUP_MODE_RICH
    from typer.rich_utils import STYLE_DEPRECATED
    from typer.rich_utils import STYLE_HELPTEXT
    from typer.rich_utils import STYLE_HELPTEXT_FIRST_LINE
    from typer.rich_utils import MarkupMode
    from typer.rich_utils import _make_rich_text

    @group()
    def _get_help_text(
        *,
        obj: Union[click.Command, click.Group],
        markup_mode: MarkupMode,
    ) -> Iterable[Union[Markdown, Text]]:
        """Build primary help text for a click command or group.

        Returns the prose help text for a command or group, rendered either as a
        Rich Text object or as Markdown.
        If the command is marked as deprecated, the deprecated string will be prepended.
        """
        # Prepend deprecated status
        if obj.deprecated:
            yield Text(DEPRECATED_STRING, style=STYLE_DEPRECATED)

        # Fetch and dedent the help text
        help_text = inspect.cleandoc(obj.help or "")

        # Trim off anything that comes after \f on its own line
        help_text = help_text.partition("\f")[0]

        # Get the first paragraph
        first_line = help_text.split("\n\n")[0]
        # Remove single linebreaks
        if markup_mode != MARKUP_MODE_MARKDOWN and not first_line.startswith("\b"):
            first_line = first_line.replace("\n", " ")
        yield _make_rich_text(
            text=first_line.strip(),
            style=STYLE_HELPTEXT_FIRST_LINE,
            markup_mode=markup_mode,
        )

        # Get remaining lines, remove single line breaks and format as dim
        remaining_paragraphs = help_text.split("\n\n")[1:]
        if remaining_paragraphs:
            if markup_mode != MARKUP_MODE_RICH:
                # Remove single linebreaks
                remaining_paragraphs = [
                    x.replace("\n", " ").strip()
                    if not x.startswith("\b")
                    else "{}\n".format(x.strip("\b\n"))
                    for x in remaining_paragraphs
                ]
                # Join back together
                remaining_lines = "\n".join(remaining_paragraphs)
            else:
                # Join with double linebreaks if markdown
                remaining_lines = "\n\n".join(remaining_paragraphs)
            # PATCH: add single newline between first and remaining lines
            yield _make_rich_text(
                text="\n",
                style=STYLE_HELPTEXT,
                markup_mode=markup_mode,
            )
            yield _make_rich_text(
                text=remaining_lines,
                style=STYLE_HELPTEXT,
                markup_mode=markup_mode,
            )

    typer.rich_utils._get_help_text = _get_help_text


def patch_generate_enum_convertor() -> None:
    """Patches enum value converter with an additional fallback to
    instantiating the enum with the value directly.
    """

    def generate_enum_convertor(enum: type[Enum]) -> Callable[[Any], Any]:
        lower_val_map = {str(val.value).lower(): val for val in enum}

        def convertor(value: Any) -> Any:
            if value is not None:
                low = str(value).lower()
                if low in lower_val_map:
                    key = lower_val_map[low]
                    return enum(key)
                # Fall back to passing in the value as-is
                try:
                    return enum(value)
                except ValueError:
                    return None

        return convertor

    typer.main.generate_enum_convertor = generate_enum_convertor


def patch_get_click_type() -> None:
    """Adds support for our custom `APIStrEnum` type.

    Used in conjunction with our custom generate_enum_convertor to support
    instantiating `APIStrEnum` with both the human-readable value and the API value
    (e.g. `"Enabled"` and `0`).

    Uses the `APIStrEnum.all_choices()` method to get the list of choices.
    """

    def get_click_type(
        *, annotation: Any, parameter_info: ParameterInfo
    ) -> click.ParamType:
        if parameter_info.click_type is not None:
            return parameter_info.click_type

        elif parameter_info.parser is not None:
            return click.types.FuncParamType(parameter_info.parser)

        elif annotation == str:  # noqa: E721
            return click.STRING
        elif annotation == int:  # noqa: E721
            if parameter_info.min is not None or parameter_info.max is not None:
                min_ = None
                max_ = None
                if parameter_info.min is not None:
                    min_ = int(parameter_info.min)
                if parameter_info.max is not None:
                    max_ = int(parameter_info.max)
                return click.IntRange(min=min_, max=max_, clamp=parameter_info.clamp)
            else:
                return click.INT
        elif annotation == float:  # noqa: E721
            if parameter_info.min is not None or parameter_info.max is not None:
                return click.FloatRange(
                    min=parameter_info.min,
                    max=parameter_info.max,
                    clamp=parameter_info.clamp,
                )
            else:
                return click.FLOAT
        elif annotation == bool:  # noqa: E721
            return click.BOOL
        elif annotation == UUID:
            return click.UUID
        elif annotation == datetime:
            return click.DateTime(formats=parameter_info.formats)
        elif (
            annotation == Path
            or parameter_info.allow_dash
            or parameter_info.path_type
            or parameter_info.resolve_path
        ):
            return click.Path(
                exists=parameter_info.exists,
                file_okay=parameter_info.file_okay,
                dir_okay=parameter_info.dir_okay,
                writable=parameter_info.writable,
                readable=parameter_info.readable,
                resolve_path=parameter_info.resolve_path,
                allow_dash=parameter_info.allow_dash,
                path_type=parameter_info.path_type,
            )
        elif lenient_issubclass(annotation, typer.FileTextWrite):
            return click.File(
                mode=parameter_info.mode or "w",
                encoding=parameter_info.encoding,
                errors=parameter_info.errors,
                lazy=parameter_info.lazy,
                atomic=parameter_info.atomic,
            )
        elif lenient_issubclass(annotation, typer.FileText):
            return click.File(
                mode=parameter_info.mode or "r",
                encoding=parameter_info.encoding,
                errors=parameter_info.errors,
                lazy=parameter_info.lazy,
                atomic=parameter_info.atomic,
            )
        elif lenient_issubclass(annotation, typer.FileBinaryRead):
            return click.File(
                mode=parameter_info.mode or "rb",
                encoding=parameter_info.encoding,
                errors=parameter_info.errors,
                lazy=parameter_info.lazy,
                atomic=parameter_info.atomic,
            )
        elif lenient_issubclass(annotation, typer.FileBinaryWrite):
            return click.File(
                mode=parameter_info.mode or "wb",
                encoding=parameter_info.encoding,
                errors=parameter_info.errors,
                lazy=parameter_info.lazy,
                atomic=parameter_info.atomic,
            )
        # our patch for APIStrEnum
        elif lenient_issubclass(annotation, APIStrEnum):
            annotation = cast(type[APIStrEnum], annotation)
            return click.Choice(
                annotation.all_choices(),
                case_sensitive=parameter_info.case_sensitive,
            )
        elif lenient_issubclass(annotation, Enum):
            return click.Choice(
                [item.value for item in annotation],
                case_sensitive=parameter_info.case_sensitive,
            )
        elif lenient_issubclass(annotation, click.Command):
            return CommandParam()

        raise RuntimeError(f"Type not yet supported: {annotation}")  # pragma no cover

    typer.main.get_click_type = get_click_type


def patch__get_rich_console() -> None:
    from rich.console import Console
    from rich.theme import Theme
    from typer.rich_utils import COLOR_SYSTEM
    from typer.rich_utils import FORCE_TERMINAL
    from typer.rich_utils import MAX_WIDTH
    from typer.rich_utils import STYLE_METAVAR
    from typer.rich_utils import STYLE_METAVAR_SEPARATOR
    from typer.rich_utils import STYLE_NEGATIVE_OPTION
    from typer.rich_utils import STYLE_NEGATIVE_SWITCH
    from typer.rich_utils import STYLE_OPTION
    from typer.rich_utils import STYLE_SWITCH
    from typer.rich_utils import STYLE_USAGE
    from typer.rich_utils import highlighter

    from zabbix_cli.output.style import RICH_THEME

    styles: dict[str, Union[str, Style]] = RICH_THEME.styles.copy()
    styles.update(
        {
            "option": STYLE_OPTION,
            "switch": STYLE_SWITCH,
            "negative_option": STYLE_NEGATIVE_OPTION,
            "negative_switch": STYLE_NEGATIVE_SWITCH,
            "metavar": STYLE_METAVAR,
            "metavar_sep": STYLE_METAVAR_SEPARATOR,
            "usage": STYLE_USAGE,
        },
    )
    TYPER_THEME = Theme(styles)

    def _get_rich_console(stderr: bool = False) -> Console:  # noqa: FBT002
        return Console(
            theme=TYPER_THEME,
            highlighter=highlighter,
            color_system=COLOR_SYSTEM,
            force_terminal=FORCE_TERMINAL,
            width=MAX_WIDTH,
            stderr=stderr,
        )

    typer.rich_utils._get_rich_console = _get_rich_console


def patch() -> None:
    """Apply all patches."""
    with patcher("typer.rich_utils.STYLE_HELPTEXT"):
        patch_help_text_style()
    with patcher("typer.rich_utils._get_help_text"):
        patch_help_text_spacing()
    with patcher("typer.main.generate_enum_convertor"):
        patch_generate_enum_convertor()
    with patcher("typer.main.get_click_type"):
        patch_get_click_type()
    with patcher("typer.rich_utils._get_rich_console"):
        patch__get_rich_console()



from __future__ import annotations

from itertools import chain
from typing import Optional

import typer

from zabbix_cli.app import Example
from zabbix_cli.app import app
from zabbix_cli.commands.common.args import ARG_HOSTNAMES_OR_IDS
from zabbix_cli.commands.common.args import ARG_TEMPLATE_NAMES_OR_IDS
from zabbix_cli.output.console import info
from zabbix_cli.output.console import success
from zabbix_cli.output.formatting.grammar import pluralize as p
from zabbix_cli.output.render import render_result
from zabbix_cli.utils.args import parse_hosts_arg
from zabbix_cli.utils.args import parse_list_arg
from zabbix_cli.utils.args import parse_templates_arg

HELP_PANEL = "Template"


@app.command(
    "link_template_to_host",
    rich_help_panel=HELP_PANEL,
    examples=[
        Example(
            "Link one template to one host",
            "link_template_to_host 'Apache by HTTP' foo.example.com",
        ),
        Example(
            "Link many templates to many hosts",
            "link_template_to_host 'Apache by HTTP,HAProxy by Zabbix agent' foo.example.com,bar.example.com",
        ),
        Example(
            "Link one template to all hosts",
            "link_template_to_host 'Apache by HTTP' '*'",
        ),
        Example(
            "Link many templates to all hosts",
            "link_template_to_host 'Apache by HTTP,HAProxy by Zabbix agent' '*'",
        ),
        Example(
            "Link all templates to all hosts [red](use with caution!)[/red]",
            "link_template_to_host '*' '*'",
        ),
    ],
)
def link_template_to_host(
    ctx: typer.Context,
    template_names_or_ids: str = ARG_TEMPLATE_NAMES_OR_IDS,
    hostnames_or_ids: str = ARG_HOSTNAMES_OR_IDS,
    strict: bool = typer.Option(
        False,
        "--strict",
        help="Fail if any hosts or templates aren't found. Should not be used in conjunction with wildcards.",
    ),
    dryrun: bool = typer.Option(
        False,
        "--dryrun",
        help="Preview changes.",
    ),
) -> None:
    """Link templates to hosts."""
    from zabbix_cli.commands.results.template import LinkTemplateToHostResult
    from zabbix_cli.models import AggregateResult

    templates = parse_templates_arg(
        app, template_names_or_ids, strict=strict, select_hosts=True
    )
    hosts = parse_hosts_arg(app, hostnames_or_ids, strict=strict)
    if not dryrun:
        with app.state.console.status("Linking templates..."):
            app.state.client.link_templates_to_hosts(templates, hosts)

    result: list[LinkTemplateToHostResult] = []
    for host in hosts:
        r = LinkTemplateToHostResult.from_result(templates, host, "Link")
        if not r.templates:
            continue
        result.append(r)

    total_templates = len(set(chain.from_iterable((r.templates) for r in result)))
    total_hosts = len(result)

    render_result(AggregateResult(result=result))
    base_msg = f"{p('template', total_templates)} to {p('host', total_hosts)}"
    if dryrun:
        info(f"Would link {base_msg}.")
    else:
        success(f"Linked {base_msg}.")


@app.command(
    "link_template_to_template",
    rich_help_panel=HELP_PANEL,
    examples=[
        Example(
            "Link one template to one template",
            "link_template_to_template 'Apache by HTTP' foo_template",
        ),
        Example(
            "Link many templates to many templates",
            "link_template_to_template 'Apache by HTTP,HAProxy by Zabbix agent' foo_template,bar_template",
        ),
        Example(
            "Link all templates starting with 'Apache' to a template",
            "link_template_to_template 'Apache*' foo_template",
        ),
        Example(
            "Link all templates containing 'HTTP' to a subset of templates",
            "link_template_to_template '*HTTP*' 'Webserver-*'",
        ),
    ],
)
def link_template_to_template(
    ctx: typer.Context,
    source: str = ARG_TEMPLATE_NAMES_OR_IDS,
    dest: str = ARG_TEMPLATE_NAMES_OR_IDS,
    strict: bool = typer.Option(
        False,
        "--strict",
        help="Fail if any templates aren't found. Should not be used in conjunction with wildcards.",
    ),
    dryrun: bool = typer.Option(
        False,
        "--dryrun",
        help="Do not actually link templates, just show what would be done.",
    ),
) -> None:
    """Link templates to templates.

    [b]NOTE:[/] Destination templates are the ones that are ultimately modified. Source templates remain unchanged.
    """
    from zabbix_cli.commands.results.template import LinkTemplateResult

    # TODO: add existing link checking just like in `link_template_to_host` & `unlink_template_from_host`
    # so we only print the ones that are actually linked
    source_templates = parse_templates_arg(app, source, strict=strict)
    dest_templates = parse_templates_arg(app, dest, strict=strict)

    if not dryrun:
        with app.state.console.status("Linking templates..."):
            app.state.client.link_templates(source_templates, dest_templates)
    render_result(
        LinkTemplateResult.from_result(source_templates, dest_templates, "Link")
    )
    base_msg = f"{p('template', len(source_templates))} to {p('template', len(dest_templates))}"
    if dryrun:
        info(f"Would link {base_msg}.")
    else:
        success(f"Linked {base_msg}.")


@app.command("show_template", rich_help_panel=HELP_PANEL)
def show_template(
    ctx: typer.Context,
    template_name: str = typer.Argument(
        help="Template name or ID. Names support wildcards.",
        show_default=False,
    ),
) -> None:
    """Show a template."""
    template = app.state.client.get_template(
        template_name,
        select_hosts=True,
        select_templates=True,
        select_parent_templates=True,
    )
    render_result(template)


@app.command("show_templates", rich_help_panel=HELP_PANEL)
def show_templates(
    ctx: typer.Context,
    templates: Optional[str] = typer.Argument(
        None,
        help="Template name(s) or ID(s). Comma-separated. Supports wildcards.",
        show_default=False,
    ),
) -> None:
    """Show all templates.

    Shows all templates by default. The template name can be a pattern containing wildcards.
    Names and IDs cannot be mixed.
    """
    from zabbix_cli.models import AggregateResult

    template_names_or_ids = parse_list_arg(templates)
    tpls = app.state.client.get_templates(
        *template_names_or_ids,
        select_hosts=True,
        select_templates=True,
        select_parent_templates=True,
    )
    render_result(AggregateResult(result=tpls))


@app.command(
    "show_items",
    rich_help_panel=HELP_PANEL,
    examples=[
        Example(
            "Show items for a template",
            "show_items 'Apache by HTTP'",
        ),
    ],
)
def show_items(
    ctx: typer.Context,
    template_name: str = typer.Argument(
        help="Template name or ID. Supports wildcards.",
        show_default=False,
    ),
) -> None:
    """Show a template's items."""
    from zabbix_cli.models import AggregateResult

    template = app.state.client.get_template(template_name)
    items = app.state.client.get_items(templates=[template])
    # NOTE: __title__ is ignored by Pydantic when used as a kwarg
    # when instantiating a model. We either need to subclass AggregateResult
    # or set it after instantiation.
    # Ideally, we would rename the field to `table_title` and make it an actual field
    # but that clashes with our existing `__<option>__` pattern for overriding
    res = AggregateResult(result=items)
    res.__title__ = template.host
    render_result(res)


@app.command(
    "unlink_template_from_host",
    rich_help_panel=HELP_PANEL,
    examples=[
        Example(
            "Unlink a template from a host",
            "unlink_template_from_host 'Apache by HTTP' foo.example.com",
        ),
        Example(
            "Unlink many templates from many hosts",
            "unlink_template_from_host 'Apache by HTTP,HAProxy by Zabbix agent' foo.example.com,bar.example.com",
        ),
        Example(
            "Unlink one template from all hosts",
            "unlink_template_from_host 'Apache by HTTP' '*'",
        ),
        Example(
            "Unlink templates starting with 'Apache' from hosts starting with 'Web'",
            "unlink_template_from_host 'Apache*' 'Web-*'",
        ),
        Example(
            "Unlink template from host without clearing items and triggers",
            "unlink_template_from_host --no-clear 'Apache by HTTP' foo.example.com",
        ),
    ],
)
def unlink_template_from_host(
    ctx: typer.Context,
    template_names_or_ids: str = ARG_TEMPLATE_NAMES_OR_IDS,
    hostnames_or_ids: str = ARG_HOSTNAMES_OR_IDS,
    strict: bool = typer.Option(
        False,
        "--strict",
        help="Fail if any hosts or templates aren't found. Should not be used in conjunction with wildcards.",
    ),
    dryrun: bool = typer.Option(
        False,
        "--dryrun",
        help="Preview changes.",
    ),
    clear: bool = typer.Option(
        True, "--clear/--no-clear", help="Unlink and clear templates."
    ),
) -> None:
    """Unlink templates from hosts.

    Unlinks and clears by default. Use `--no-clear` to unlink without clearing.
    """
    from zabbix_cli.commands.results.template import UnlinkTemplateFromHostResult
    from zabbix_cli.models import AggregateResult

    templates = parse_templates_arg(
        app, template_names_or_ids, strict=strict, select_hosts=True
    )
    hosts = parse_hosts_arg(app, hostnames_or_ids, strict=strict)

    action = "Unlink and clear" if clear else "Unlink"
    if not dryrun:
        with app.state.console.status("Unlinking templates..."):
            app.state.client.unlink_templates_from_hosts(templates, hosts)

    # Only show hosts with matching templates to unlink
    result: list[UnlinkTemplateFromHostResult] = []
    for host in hosts:
        r = UnlinkTemplateFromHostResult.from_result(templates, host, action)
        if not r.templates:
            continue
        result.append(r)

    total_templates = len(set(chain.from_iterable((r.templates) for r in result)))
    total_hosts = len(result)

    render_result(AggregateResult(result=result))
    base_msg = f"{p('template', total_templates)} from {p('host', total_hosts)}"
    if dryrun:
        info(f"Would {action.lower()} {base_msg}.")
    else:
        action_success = "Unlinked and cleared" if clear else "Unlinked"
        success(f"{action_success} {base_msg}.")


@app.command(
    "unlink_template_from_template",
    rich_help_panel=HELP_PANEL,
    examples=[
        Example(
            "Unlink one template from one template",
            "unlink_template_from_template 'Apache by HTTP' foo_template",
        ),
        Example(
            "Unlink many templates from many templates",
            "unlink_template_from_template 'Apache by HTTP,HAProxy by Zabbix agent' foo_template,bar_template",
        ),
        Example(
            "Unlink all templates starting with 'Apache' from a template",
            "unlink_template_from_template 'Apache*' foo_template",
        ),
        Example(
            "Unlink all templates containing 'HTTP' from a subset of templates",
            "unlink_template_from_template '*HTTP*' 'Web-*'",
        ),
        Example(
            "Unlink a template without clearing items and triggers",
            "unlink_template_from_template --no-clear foo_template bar_template",
        ),
    ],
)
def unlink_template_from_template(
    ctx: typer.Context,
    source: str = ARG_TEMPLATE_NAMES_OR_IDS,
    dest: str = ARG_TEMPLATE_NAMES_OR_IDS,
    strict: bool = typer.Option(
        False,
        "--strict",
        help="Fail if any templates aren't found. Should not be used in conjunction with wildcards.",
    ),
    clear: bool = typer.Option(
        True,
        "--clear/--no-clear",
        help="Unlink and clear templates.",
    ),
    dryrun: bool = typer.Option(
        False,
        "--dryrun",
        help="Preview changes.",
    ),
) -> None:
    """Unlink templates from templates.

    Unlinks and clears by default. Use `--no-clear` to unlink without clearing.
    [b]NOTE:[/] Destination templates are the ones that are ultimately modified. Source templates remain unchanged.
    """
    from zabbix_cli.commands.results.template import LinkTemplateResult

    source_templates = parse_templates_arg(app, source, strict=strict)
    dest_templates = parse_templates_arg(app, dest, strict=strict)
    if not dryrun:
        with app.state.console.status("Unlinking templates..."):
            app.state.client.unlink_templates(
                source_templates, dest_templates, clear=clear
            )
    action = "Unlink and clear" if clear else "Unlink"
    render_result(
        LinkTemplateResult.from_result(
            source_templates,
            dest_templates,
            action,
        )
    )
    base_msg = f"{p('template', len(source_templates))} from {p('template', len(dest_templates))}"
    if dryrun:
        info(f"Would {action.lower()} {base_msg}.")
    else:
        action_success = "Unlinked and cleared" if clear else "Unlinked"
        success(f"{action_success} {base_msg}.")



from __future__ import annotations

import os
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any
from typing import NoReturn
from typing import Optional

import typer
from rich.console import Console

from zabbix_cli.logs import logger
from zabbix_cli.output.formatting.path import path_link
from zabbix_cli.output.style import RICH_THEME
from zabbix_cli.output.style import Icon
from zabbix_cli.state import get_state

if TYPE_CHECKING:
    from rich.theme import Theme

    from zabbix_cli.config.model import Config

# stdout console used to print results
console = Console(theme=RICH_THEME)

# stderr console used to print prompts, messages, etc.
err_console = Console(
    stderr=True,
    highlight=False,
    soft_wrap=True,
    theme=RICH_THEME,
)


def get_theme(name: str) -> Theme:
    """Get a Rich console theme by name."""
    return RICH_THEME  # other themes NYI


RESERVED_EXTRA_KEYS = (
    "name",
    "level",
    "pathname",
    "lineno",
    "msg",
    "args",
    "exc_info",
    "func",
    "sinfo",
)


def get_extra_dict(**kwargs: Any) -> dict[str, Any]:
    """Format the extra dict for logging. Renames some keys to avoid
    collisions with the default keys.

    See: https://docs.python.org/3.11/library/logging.html#logging.LogRecord
    """
    for k in list(kwargs):  # iterate over copy while mutating
        if k in RESERVED_EXTRA_KEYS:
            kwargs[f"{k}_"] = kwargs.pop(k)
    return kwargs


def debug_kv(key: str, value: Any) -> None:
    """Print and log a key value pair."""
    msg = f"[bold]{key:<20}:[/bold] {value}"
    from rich.markup import render

    logger.debug(
        render(msg).plain, extra=get_extra_dict(key=key, value=value), stacklevel=2
    )
    err_console.print(msg)


def debug(message: str, icon: str = "", *args: Any, **kwargs: Any) -> None:
    """Log with INFO level and print an informational message."""
    logger.debug(message, extra=get_extra_dict(**kwargs), stacklevel=2)
    err_console.print(message)


def info(message: str, icon: str = Icon.INFO, *args: Any, **kwargs: Any) -> None:
    """Log with INFO level and print an informational message."""
    logger.info(message, extra=get_extra_dict(**kwargs), stacklevel=2)
    err_console.print(f"[success]{icon}[/] {message}")


def success(message: str, icon: str = Icon.OK, **kwargs: Any) -> None:
    """Log with INFO level and print a success message."""
    logger.info(message, extra=get_extra_dict(**kwargs), stacklevel=2)
    err_console.print(f"[success]{icon}[/] {message}")


def warning(message: str, icon: str = Icon.WARNING, **kwargs: Any) -> None:
    """Log with WARNING level and optionally print a warning message."""
    logger.warning(message, extra=get_extra_dict(**kwargs), stacklevel=2)
    err_console.print(f"[warning]{icon} {message}[/]")


def error(
    message: str,
    icon: str = Icon.ERROR,
    *,
    exc_info: bool = False,
    log: bool = True,
    **kwargs: Any,
) -> None:
    """Log with ERROR level and print an error message."""
    if log:  # we can disable logging when the logger isn't set up yet
        logger.error(
            message, extra=get_extra_dict(**kwargs), exc_info=exc_info, stacklevel=2
        )
    err_console.print(f"[error]{icon} ERROR: {message}")


def print_help(ctx: typer.Context) -> None:
    err_console.print(ctx.get_help())


def exit_ok(message: Optional[str] = None, code: int = 0, **kwargs: Any) -> NoReturn:
    """Logs a message with INFO level and exits with the given code (default: 0)

    Parameters
    ----------
    message : str
        Message to print.
    code : int, optional
        Exit code, by default 0
    **kwargs
        Additional keyword arguments to pass to the extra dict.
    """
    if message:
        info(message, **kwargs)
    raise SystemExit(code)


def exit_err(
    message: str, code: int = 1, exception: Optional[Exception] = None, **kwargs: Any
) -> NoReturn:
    """Logs a message with ERROR level and exits with the given
    code (default: 1).

    Parameters
    ----------
    message : str
        Message to print.
    code : int, optional
        Exit code, by default 1
    **kwargs
        Additional keyword arguments to pass to the extra dict.
    """
    state = get_state()

    # Render JSON-formatted error message if output format is JSON
    if state.config.app.output.format == "json":
        from zabbix_cli.exceptions import get_cause_args
        from zabbix_cli.models import Result
        from zabbix_cli.models import ReturnCode
        from zabbix_cli.output.render import render_json

        errors = get_cause_args(exception)
        render_json(
            Result(message=message, return_code=ReturnCode.ERROR, errors=errors)
        )
    else:
        error(message, **kwargs)
    raise SystemExit(code)


def print_toml(toml_str: str) -> None:
    """Prints TOML to stdout using the default console."""
    console.print(
        toml_str,
        markup=False,  # TOML tables could be interpreted as rich markup
        soft_wrap=True,  # prevents mangling whitespace
    )


def print_path(path: Path) -> None:
    """Prints a path to stdout using the default console."""
    console.print(
        path_link(path),
        highlight=False,
        soft_wrap=True,
    )


def disable_color() -> None:
    """Disable color output in consoles."""
    console._color_system = None  # pyright: ignore[reportPrivateUsage]
    err_console._color_system = None  # pyright: ignore[reportPrivateUsage]
    # HACK: set env var to disable color in Typer console
    os.environ["NO_COLOR"] = "1"


def enable_color() -> None:
    """Enable color output in consoles if supported."""
    console._color_system = console._detect_color_system()  # pyright: ignore[reportPrivateUsage]
    err_console._color_system = console._detect_color_system()  # pyright: ignore[reportPrivateUsage]
    os.unsetenv("NO_COLOR")  # remove hack


def configure_console(config: Config) -> None:
    """Configure console output based on the application configuration."""
    if not config.app.output.color:
        disable_color()



from __future__ import annotations

from typing import Any
from typing import Literal

import pytest
from inline_snapshot import snapshot
from packaging.version import Version
from pytest_httpserver import HTTPServer
from zabbix_cli.exceptions import ZabbixAPILoginError
from zabbix_cli.exceptions import ZabbixAPILogoutError
from zabbix_cli.pyzabbix.client import ZabbixAPI
from zabbix_cli.pyzabbix.client import add_param
from zabbix_cli.pyzabbix.client import append_param

from tests.utils import add_zabbix_endpoint
from tests.utils import add_zabbix_version_endpoint


@pytest.mark.parametrize(
    "inp, key, value, expect",
    [
        pytest.param(
            {"hostids": 1},
            "hostids",
            2,
            {"hostids": [1, 2]},
            id="non-list (int)",
        ),
        pytest.param(
            {"hostids": [1]},
            "hostids",
            2,
            {"hostids": [1, 2]},
            id="list (int)",
        ),
        pytest.param(
            {"hostids": "1"},
            "hostids",
            "2",
            {"hostids": ["1", "2"]},
            id="non-list (str)",
        ),
        pytest.param(
            {"hostids": ["1"]},
            "hostids",
            "2",
            {"hostids": ["1", "2"]},
            id="list (str)",
        ),
        pytest.param(
            {"lists": [[1, 2, 3]]},
            "lists",
            [4, 5, 6],
            {"lists": [[1, 2, 3], [4, 5, 6]]},
            id="list of lists",
        ),
    ],
)
def test_append_param(inp: Any, key: str, value: Any, expect: dict[str, Any]) -> None:
    result = append_param(inp, key, value)
    assert result == expect
    # Check in-place modification
    assert result is inp


@pytest.mark.parametrize(
    "inp, subkey, value, expect",
    [
        (
            {"output": "extend"},
            "hostids",
            2,
            {"output": "extend", "search": {"hostids": 2}},
        ),
    ],
)
def test_add_param(inp: Any, subkey: str, value: Any, expect: dict[str, Any]) -> None:
    result = add_param(inp, "search", subkey, value)
    assert result == expect
    # Check in-place modification


def test_login_fails(zabbix_client_mock_version: ZabbixAPI) -> None:
    client = zabbix_client_mock_version
    client.set_url("http://some-url-that-will-fail.gg")
    assert client.url == snapshot("http://some-url-that-will-fail.gg/api_jsonrpc.php")

    with pytest.raises(ZabbixAPILoginError) as exc_info:
        client.login(user="username", password="password")

    assert exc_info.exconly() == snapshot(
        "zabbix_cli.exceptions.ZabbixAPILoginError: Failed to log in to Zabbix: Failed to send request to http://some-url-that-will-fail.gg/api_jsonrpc.php (user.login) with params {'username': 'username', 'password': 'password'}"
    )
    assert str(exc_info.value) == snapshot(
        "Failed to log in to Zabbix: Failed to send request to http://some-url-that-will-fail.gg/api_jsonrpc.php (user.login) with params {'username': 'username', 'password': 'password'}"
    )


def test_logout_fails(zabbix_client_mock_version: ZabbixAPI) -> None:
    """Test that we get the correct exception type when login fails
    due to a connection error.
    """
    client = zabbix_client_mock_version
    client.set_url("http://some-url-that-will-fail.gg")
    assert client.url == snapshot("http://some-url-that-will-fail.gg/api_jsonrpc.php")

    client.auth = "authtoken123456789"

    with pytest.raises(ZabbixAPILogoutError) as exc_info:
        client.logout()

    assert exc_info.exconly() == snapshot(
        "zabbix_cli.exceptions.ZabbixAPILogoutError: Failed to log out of Zabbix: Failed to send request to http://some-url-that-will-fail.gg/api_jsonrpc.php (user.logout) with params {}"
    )


@pytest.mark.parametrize(
    "inp, expect",
    [
        pytest.param(
            "http://localhost",
            "http://localhost/api_jsonrpc.php",
            id="localhost-no-slash",
        ),
        pytest.param(
            "http://localhost/",
            "http://localhost/api_jsonrpc.php",
            id="localhost-with-slash",
        ),
        pytest.param(
            "http://localhost/api_jsonrpc.php",
            "http://localhost/api_jsonrpc.php",
            id="localhost-full-url",
        ),
        pytest.param(
            "http://localhost/api_jsonrpc.php/",
            "http://localhost/api_jsonrpc.php",
            id="localhost-full-url-with-slash",
        ),
        pytest.param(
            "http://example.com",
            "http://example.com/api_jsonrpc.php",
            id="tld-no-slash",
        ),
        pytest.param(
            "http://example.com/",
            "http://example.com/api_jsonrpc.php",
            id="tld-with-slash",
        ),
        pytest.param(
            "http://example.com/api_jsonrpc.php",
            "http://example.com/api_jsonrpc.php",
            id="tld-full-url",
        ),
        pytest.param(
            "http://example.com/api_jsonrpc.php/",
            "http://example.com/api_jsonrpc.php",
            id="tld-full-url-with-slash",
        ),
    ],
)
def test_client_server_url(inp: str, expect: str) -> None:
    zabbix_client = ZabbixAPI(server=inp)
    assert zabbix_client.url == expect


AuthMethod = Literal["header", "body"]


@pytest.mark.parametrize(
    "version,expect_method",
    [
        pytest.param(Version("5.0.0"), "body", id="5.0.0"),
        pytest.param(Version("5.2.0"), "body", id="5.2.0"),
        pytest.param(Version("6.0.0"), "body", id="6.0.0"),
        pytest.param(Version("6.2.0"), "body", id="6.2.0"),
        pytest.param(Version("6.4.0"), "header", id="6.4.0"),
        pytest.param(Version("7.0.0"), "header", id="7.0.0"),
        pytest.param(Version("7.2.0"), "header", id="7.2.0"),
    ],
)
def test_client_auth_method(
    zabbix_client: ZabbixAPI,
    httpserver: HTTPServer,
    version: Version,
    expect_method: AuthMethod,
) -> None:
    """Test that the correct auth method (body/header) is used based on the Zabbix server version."""
    # Add endpoint for version check
    zabbix_client.set_url(httpserver.url_for("/api_jsonrpc.php"))

    # Set a mock token we can use for testing
    zabbix_client.auth = "token123"

    # Add endpoint that returns the parametrized version
    add_zabbix_version_endpoint(httpserver, str(version), id=0)

    assert zabbix_client.version == version

    headers: dict[str, str] = {}
    auth = None
    # We expect auth token to be in header on >= 6.4.0
    if expect_method == "header":
        headers["Authorization"] = f"Bearer {zabbix_client.auth}"
    else:
        auth = zabbix_client.auth

    add_zabbix_endpoint(
        httpserver,
        method="test.method.do_stuff",
        params={},
        response="authtoken123456789",
        headers=headers,
        auth=auth,
    )

    # Will fail if the auth method is not set correctly
    resp = zabbix_client.do_request("test.method.do_stuff")
    assert resp.result == "authtoken123456789"

    httpserver.check_assertions()
    httpserver.check_handler_errors()


AuthType = Literal["token", "sessionid"]


@pytest.mark.parametrize(
    "auth_type,auth",
    [
        pytest.param("token", "authtoken123456789", id="token"),
        pytest.param("token", "", id="token (empty string)"),
        pytest.param("sessionid", "sessionid123456789", id="sessionid"),
        pytest.param("sessionid", "", id="sessionid (empty string)"),
    ],
)
def test_client_logout(httpserver: HTTPServer, auth_type: AuthType, auth: str) -> None:
    add_zabbix_version_endpoint(httpserver, "7.0.0")

    # We only expect a logout request if we are using a sessionid and have an auth token
    if auth_type == "sessionid" and auth:
        add_zabbix_endpoint(
            httpserver,
            "user.logout",
            params={},
            response=True,  # the value `True` as the response
        )
    zabbix_client = ZabbixAPI(server=httpserver.url_for("/api_jsonrpc.php"))
    zabbix_client.auth = auth
    if auth_type == "token":
        zabbix_client.use_api_token = True
    zabbix_client.logout()

    httpserver.check_assertions()
    httpserver.check_handler_errors()



"""Config wizard for Zabbix CLI."""

from __future__ import annotations

import ast
import inspect
import logging
import sys
import textwrap
from collections.abc import Iterable
from dataclasses import dataclass
from dataclasses import field
from enum import Enum
from enum import EnumMeta
from operator import attrgetter
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Generic
from typing import Optional
from typing import TypeVar
from typing import Union
from typing import get_args

from pydantic import SecretStr
from rich.panel import Panel

from zabbix_cli.bulk import BulkRunnerMode
from zabbix_cli.config.constants import OutputFormat
from zabbix_cli.config.model import Config
from zabbix_cli.exceptions import ZabbixCLIError
from zabbix_cli.output.console import console
from zabbix_cli.output.console import exit_err
from zabbix_cli.output.prompts import bool_prompt
from zabbix_cli.output.prompts import float_prompt
from zabbix_cli.output.prompts import int_prompt
from zabbix_cli.output.prompts import list_prompt
from zabbix_cli.output.prompts import path_prompt
from zabbix_cli.output.prompts import str_prompt
from zabbix_cli.utils.args import parse_bool_arg

if sys.version_info >= (3, 10):
    from itertools import pairwise
else:
    # source: https://docs.python.org/3.12/library/itertools.html#itertools.pairwise
    def pairwise(iterable: Iterable[Any]) -> Iterable[tuple[Any, Any]]:
        # pairwise('ABCDEFG') → AB BC CD DE EF FG
        iterator = iter(iterable)
        a = next(iterator, None)
        for b in iterator:
            yield a, b
            a = b


logger = logging.getLogger(__name__)


EnumT = TypeVar("EnumT", bound=Enum)


@dataclass
class EnumMember(Generic[EnumT]):
    member: EnumT
    description: str = ""  # docstring


# Adapted from https://davidism.com/attribute-docstrings/
def get_enum_attr_docs(cls: type[EnumT]) -> dict[EnumT, str]:
    """Get any docstrings placed after attribute assignments in a class body."""
    cls_node = ast.parse(textwrap.dedent(inspect.getsource(cls))).body[0]

    if not isinstance(cls_node, ast.ClassDef):
        raise TypeError("Given object was not a class.")

    out: dict[EnumT, str] = {}

    # Consider each pair of nodes.
    for a, b in pairwise(cls_node.body):
        # Must be an assignment then a constant string.
        if (
            not isinstance(a, (ast.Assign, ast.AnnAssign))
            or not isinstance(b, ast.Expr)
            or not isinstance(b.value, ast.Constant)
            or not isinstance(b.value.value, str)
        ):
            continue

        doc = inspect.cleandoc(b.value.value)

        if isinstance(a, ast.Assign):
            # An assignment can have multiple targets (a = b = v).
            targets = a.targets
        else:
            # An annotated assignment only has one target.
            targets = [a.target]

        for target in targets:
            # Must be assigning to a plain name.
            if not isinstance(target, ast.Name):
                continue

            out[getattr(cls, target.id)] = doc

    return out


def get_enum_members(enum: type[EnumT]) -> list[EnumMember[EnumT]]:
    """Get enum members as a list of EnumMember namedtuples."""
    # Get the docstrings for the enum members
    try:
        docs = get_enum_attr_docs(enum)
    except Exception as e:
        logger.error("Failed to get attribute docstrings for enum %s: %s", enum, e)
        docs: dict[Enum, str] = {}
    return [
        EnumMember(member=member, description=docs.get(member, "")) for member in enum
    ]


def enum_prompt(
    choice_type: str,
    choices: type[EnumT],
    *,
    default: EnumT | None = None,
    show_default: bool = False,
    **kwargs: Any,
) -> EnumT:
    """Prompt for a choice input given an enum."""
    members = get_enum_members(choices)
    member_map = {str(idx): member for idx, member in enumerate(members, start=1)}  # noqa: C416 # dict comprehension is clearer
    lines = [choice_type]
    for idx, member in member_map.items():
        # TODO: add consistent formatting of spacing
        lines.append(f"{idx}) {member.member.name} - {member.description}")
    console.print("\n".join(lines))

    if default:
        default_arg = next(
            iter(k for k, v in member_map.items() if v.member == default), None
        )
    else:
        default_arg = None

    choice = str_prompt(
        "Enter selection",
        default=default_arg,
        show_default=show_default,
        choices=list(member_map.keys()),
        show_choices=True,
        **kwargs,
    )
    try:
        return member_map[choice].member
    except KeyError as e:
        logger.error("Failed to get choice %s for enum %s: %s", choice, choices, e)
        # NOTE: not really sure how to write this error message.
        # In theory it should be unreachable.
        raise ValueError(
            f"Unexpected choice {choice} for {choice_type} - expected one of {list(member_map.keys())}"
        ) from e
    except ValueError as e:
        logger.error("Unable to convert choice %s to %s enum: %s", choice, choices, e)
        raise ValueError(f"Unable to parse choice {choice!r} for {choice_type}") from e


def is_enum_type(obj: object) -> bool:
    """Check if an object is an enum."""
    return type(obj) is EnumMeta


T = TypeVar("T")


class AnySet(set[T]):
    pass


class AllSet(set[T]):
    pass


# TODO: NotAnySet
# TODO: NotAllSet


@dataclass
class ConfigOption(Generic[T]):
    """Configuration option for the CLI."""

    name: str
    attr: str
    type: type[T]
    empty_ok: bool = True
    message: Optional[str] = None
    depends_on: Union[AnySet[str], AllSet[str]] = field(default_factory=AnySet)

    def get_value(self, config: Config) -> T:
        return attrgetter(self.attr)(config)

    def get_message(self) -> str:
        """Get the message for the prompt."""
        if self.message:
            return self.message
        elif is_enum_type(self.type):
            return f"Select {self.name}"
        return f"Enter {self.name}"


def secret_str_prompt(
    message: str,
    *,
    default: SecretStr | None = None,
    show_default: bool = False,
    empty_ok: bool = False,
    password: bool = True,
) -> SecretStr:
    """Prompt for a secret string input."""
    value = str_prompt(
        message,
        default=default.get_secret_value() if default else None,
        show_default=show_default,
        empty_ok=empty_ok,
        password=password,
    )
    return SecretStr(value)


PROMPT_MAP: dict[type, Callable[..., object]] = {
    str: str_prompt,
    int: str_prompt,
    float: float_prompt,
    bool: bool_prompt,
    list: list_prompt,
    SecretStr: secret_str_prompt,
    Enum: enum_prompt,
}


def get_prompt(type: type[T]) -> Callable[..., object]:
    """Get the prompt function for a given type."""
    if type in PROMPT_MAP:
        return PROMPT_MAP[type]
    elif issubclass(type, Enum):
        return enum_prompt
    raise ValueError(f"Unsupported type for config option: {type}")


def bool_path_prompt(
    message: str,
    attr: str,
    *,
    default: Union[bool, Path] = True,
    show_default: bool = False,
    empty_ok: bool = False,
) -> Union[bool, Path]:
    """Prompt for a boolean or path input."""
    default_val = str(default) if isinstance(default, Path) else "y" if default else "n"
    inp = str_prompt(
        f"{message} [y/n/<path>]",
        default=default_val,
        show_default=True,
        show_choices=False,
    )
    try:
        val = parse_bool_arg(inp)
    except ZabbixCLIError:
        try:
            val = Path(inp)
        except Exception:
            exit_err(
                f"Invalid input for {attr}: {inp}. Must be 'y', 'n', or a valid path to a file."
            )
    return val


_CONFIG_OPTIONS: dict[str, list[ConfigOption[Any]]] = {
    "API Connection Settings": [
        ConfigOption(
            name="API URL",
            attr="api.url",
            type=str,
        ),
        ConfigOption(
            name="Username",
            attr="api.username",
            type=str,
        ),
        ConfigOption(
            name="Password",
            attr="api.password",
            type=SecretStr,
        ),
        ConfigOption(
            name="Auth token",
            attr="api.auth_token",
            type=SecretStr,
        ),
        ConfigOption(
            name="Verify SSL",
            message="Verify SSL certificates? (y/n or path to custom CA bundle)",
            attr="api.verify_ssl",
            type=Union[bool, Path],  # pyright: ignore[reportArgumentType] # TODO: fix union types
        ),
    ],
    "Application Settings": [
        ConfigOption(
            name="Session file",
            message="Store API session IDs locally to avoid login prompt each time?",
            attr="app.use_session_file",
            type=bool,
        ),
        ConfigOption(
            name="Session file location",
            attr="app.session_file",
            type=Path,
            depends_on=AnySet({"app.use_session_file"}),
        ),
        ConfigOption(
            name="Auth file",
            message="Configure auth file? (`username::password` stored in plaintext)",
            attr="app.use_auth_file",
            type=bool,
        ),
        ConfigOption(
            name="Auth file location",
            attr="app.auth_file",
            type=Path,
            depends_on=AnySet({"app.use_auth_file"}),
        ),
        ConfigOption(
            name="Insecure session/auth file",
            message="Allow insecure session/auth file",
            attr="app.allow_insecure_auth_file",
            type=bool,
            depends_on=AnySet({"app.use_session_file", "app.auth_file"}),
        ),
        ConfigOption(
            name="Enable command history",
            message="Enable command history",
            attr="app.history",
            type=bool,
        ),
        ConfigOption(
            name="Command history file location",
            attr="app.history_file",
            type=Path,
            depends_on=AnySet({"app.history"}),
        ),
        ConfigOption(
            name="Bulk operation mode",
            attr="app.bulk_mode",
            type=BulkRunnerMode,
        ),
    ],
    "Output Settings": [
        ConfigOption(
            name="Output format",
            attr="app.output.format",
            type=OutputFormat,
        ),
        ConfigOption(
            name="Output color",
            message="Enable color output?",
            attr="app.output.color",
            type=bool,
        ),
        # paging NYI
    ],
}

COMMAND_OPTIONS = {
    # Hosts
    "create_host": [
        ConfigOption(
            name="Default host groups",
            attr="app.commands.create_host.hostgroups",
            type=list[str],
        ),
        ConfigOption(
            name="Interface creation",
            message="Automatically create interfaces?",
            attr="app.commands.create_host.create_interface",
            type=bool,
        ),
    ],
    # Groups
    "create_hostgroup": [
        ConfigOption(
            name="Default read-only user groups",
            attr="app.commands.create_hostgroup.ro_groups",
            type=list[str],
            empty_ok=True,
        ),
        ConfigOption(
            name="Default read/write user groups",
            attr="app.commands.create_hostgroup.rw_groups",
            type=list[str],
            empty_ok=True,
        ),
    ],
    "create_templategroup": [
        ConfigOption(
            name="Default read-only user groups",
            attr="app.commands.create_templategroup.ro_groups",
            type=list[str],
            empty_ok=True,
        ),
        ConfigOption(
            name="Default read/write user groups",
            attr="app.commands.create_templategroup.rw_groups",
            type=list[str],
            empty_ok=True,
        ),
    ],
    # Users
    "create_user": [
        ConfigOption(
            name="Default usergroups",
            attr="app.commands.create_user.usergroups",
            type=list[str],
            empty_ok=True,
        ),
    ],
    "create_notification_user": [
        ConfigOption(
            name="Default usergroups",
            attr="app.commands.create_notification_user.usergroups",
            type=list[str],
            empty_ok=True,
        ),
    ],
}


class ConfigWizard:
    def __init__(self, config: Config) -> None:
        self.config = config

    def print_section(self, title: str) -> None:
        """Print a section title."""
        console.rule(title)

    def print_command_subsection(self, command: str) -> None:
        """Print a command subsection title."""
        console.rule(f"[command]{command}[/]")

    def run(self) -> None:
        """Run the configuration wizard.

        Modifies the config object in place.
        """
        panel = Panel(
            """\
Welcome to the Zabbix CLI configuration wizard!
This wizard will help you set up your Zabbix CLI configuration file.\
""",
            title="Zabbix CLI Configuration Wizard",
            expand=False,
        )
        console.print(panel)

        str_prompt("Press Enter to continue...", empty_ok=True, show_default=False)
        self.do_run()

    def do_run(self) -> None:
        # Keep track of options we have set and their values
        options_set: dict[str, Any] = {}

        def iter_config_options(options: list[ConfigOption[Any]]) -> None:
            for option in options:
                # Option(s) must be set and not falsey
                # i.e. `use_session_file = False` will not prompt
                # for the session file location
                if option.depends_on:
                    operator = all if isinstance(option.depends_on, AllSet) else any
                    if not operator(options_set.get(dep) for dep in option.depends_on):
                        continue

                # Store the value before prompting
                v_pre = option.get_value(self.config)
                v_new = self.prompt(option)

                # Only update config if the value has changed
                if v_pre != v_new:
                    self.assign_config_value(option.attr, v_new)

                options_set[option.attr] = v_new

                console.print("")  # newline between inputs

        for section, options_or_subsubsection in _CONFIG_OPTIONS.items():
            self.print_section(section)
            iter_config_options(options_or_subsubsection)

        if bool_prompt("Configure command defaults?", default=True):
            for command, options_or_subsubsection in COMMAND_OPTIONS.items():
                self.print_command_subsection(command)
                iter_config_options(options_or_subsubsection)

    def assign_config_value(self, attr: str, value: Any) -> None:
        """Assign the value to the config object."""
        attrs = attr.split(".")
        obj = self.config

        # Decompose the attribute string into its components
        for attr in attrs[:-1]:
            if not hasattr(obj, attr):
                raise ValueError(f"Invalid config attribute: {attr}")
            obj = getattr(obj, attr)

        # Set the value on the last attribute
        if hasattr(obj, attrs[-1]):
            setattr(obj, attrs[-1], value)
        else:
            raise ValueError(f"Invalid config attribute: {attrs[-1]}")

    # TODO: fix this generic annotation.
    #       How can we convince type checker that the prompt function
    #       returns the correct type?
    def prompt(self, option: ConfigOption[Any]) -> Any:
        """Prompt for a configuration option and return the value."""
        if option.type is str:
            value = str_prompt(
                option.get_message(),
                default=option.get_value(self.config),
                show_default=True,
                empty_ok=option.empty_ok,
            )
        elif option.type == SecretStr:
            value = secret_str_prompt(
                option.get_message(),
                default=option.get_value(self.config),
                show_default=True,
                empty_ok=True,
                password=True,
            )
        elif option.type is bool:
            value = bool_prompt(
                option.get_message(),
                default=option.get_value(self.config),
                show_default=True,
            )
        elif option.type is int:
            value = int_prompt(
                option.get_message(),
                default=option.get_value(self.config),
                show_default=True,
                empty_ok=option.empty_ok,
            )
        elif option.type is float:
            value = float_prompt(
                option.get_message(),
                default=option.get_value(self.config),
                show_default=True,
                empty_ok=option.empty_ok,
            )
        elif option.type == list[str]:
            value = list_prompt(
                option.get_message(),
                default=option.get_value(self.config),
                empty_ok=option.empty_ok,
                type=str,
            )
        elif option.type == Path:
            value = path_prompt(
                option.get_message(),
                default=option.get_value(self.config),
                show_default=True,
            )
        # TODO: improve heuristic for union types
        elif (args := get_args(option.type)) and all(
            arg in (bool, Path) for arg in args
        ):
            value = bool_path_prompt(
                option.get_message(),
                option.attr,
                default=option.get_value(self.config),
                show_default=True,
                empty_ok=False,
            )
        elif is_enum_type(option.type):
            value = enum_prompt(
                option.get_message(),
                option.type,
                default=option.get_value(self.config),
                show_default=True,
            )
        else:
            raise ValueError(f"Unsupported type for config option: {option.type}")
        return value


def run_wizard(config: Config) -> None:
    """Run the Zabbix CLI configuration wizard."""
    wizard = ConfigWizard(config)
    wizard.run()



"""Module for running commands in bulk from a file.

Uses a very rudimentary parser to parse commands from a file, then
passes them to typer.Context.invoke() to run them.
"""

from __future__ import annotations

import logging
import shlex
from collections import Counter
from contextlib import contextmanager
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Optional

import typer
from pydantic import BaseModel
from pydantic import Field
from typing_extensions import Self

from zabbix_cli.config.constants import BulkRunnerMode
from zabbix_cli.exceptions import CommandFileError
from zabbix_cli.output.console import warning
from zabbix_cli.state import get_state
from zabbix_cli.utils.fs import read_file

logger = logging.getLogger(__name__)


class LineParseError(CommandFileError):
    """Line cannot be parsed."""


# NOTE: these exceptions are used for control flow only
class SkippableLine(LineParseError):
    """Line will be skipped during parsing."""


class EmptyLine(SkippableLine):
    """Line is empty."""


class CommentLine(SkippableLine):
    """Line is a comment."""


class BulkCommand(BaseModel):
    """A command to be run in bulk."""

    args: list[str] = Field(default_factory=list)
    line: str = ""  # original line from file
    line_number: int = 0

    def __str__(self) -> str:
        if self.line:
            return self.line
        return " ".join(self.args)

    @classmethod
    def from_line(cls, line: str, line_number: int = 0) -> Self:
        """Parse a command line into a BulkCommand."""
        # Early returns for empty lines and comments
        line = line.strip()
        if not line:
            raise EmptyLine("Cannot parse empty line")
        if line.startswith("#"):
            raise CommentLine("Cannot parse comment line")

        # Split the line into tokens, handling quotes and comments
        args = shlex.split(line, comments=True)
        if not args:
            raise LineParseError("No command specified")
        return cls(args=args, line=line, line_number=line_number)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self}>"


class CommandResult(Enum):
    """Result of a command execution."""

    SUCCESS = "success"
    FAILURE = "failure"
    SKIPPED = "skipped"


@dataclass
class CommandExecution:
    """Represents the execution of a single command."""

    command: BulkCommand
    result: CommandResult
    error: Optional[BaseException] = None
    line_number: Optional[int] = None


class BulkRunner:
    def __init__(
        self,
        ctx: typer.Context,
        file: Path,
        mode: BulkRunnerMode = BulkRunnerMode.STRICT,
    ) -> None:
        self.ctx = ctx
        self.file = file
        self.mode = mode
        self.executions: list[CommandExecution] = []
        """Commands that were executed."""
        self.skipped: list[CommandExecution] = []
        """Lines that were skipped during parsing."""

    @contextmanager
    def _command_context(self, command: BulkCommand):
        """Context manager for command execution with proper error handling."""

        def add_success() -> None:
            self.executions.append(
                CommandExecution(
                    command, CommandResult.SUCCESS, line_number=command.line_number
                )
            )
            logger.info("Command succeeded: %s", command)

        def add_failure(e: BaseException) -> None:
            self.executions.append(
                CommandExecution(
                    command,
                    CommandResult.FAILURE,
                    error=e,
                    line_number=command.line_number,
                )
            )
            if self.mode == BulkRunnerMode.STRICT:
                raise CommandFileError(
                    f"Command failed: [command]{command}[/]: {e}"
                ) from e
            else:
                logger.error("Command failed: %s - %s", command, e)

        try:
            yield
            add_success()
        except (SystemExit, typer.Exit) as e:
            # If we get return code 0 on an exit, we consider it a success
            code = e.code if isinstance(e, SystemExit) else e.exit_code
            if code == 0:
                add_success()
            else:
                add_failure(e)
        except Exception as e:
            add_failure(e)

    def run_bulk(self) -> Counter[CommandResult]:
        """Run commands in bulk from a file where each line is a CLI command.

        Returns:
            Dict[CommandResult, int]: Count of commands by result status
        Raises:
            CommandFileError: If command file cannot be parsed or a command fails (in STRICT mode)

        Example:
        ```bash
        $ cat /tmp/commands.txt

        # We can add comments and empty lines

        # We can use old-form positional arguments
        create_host test000001.example.net All-manual-hosts .+ 1

        # Or new-form keyword arguments
        create_host test000002.example.net --hostgroup All-manual-hosts --proxy .+ --status 1
        """
        # top-level Click command group for the application
        # Contains all commands defined via @app.command()
        group = self.ctx.command

        commands = self.load_command_file()
        for command in commands:
            with group.make_context(None, command.args, parent=self.ctx) as ctx:
                with self._command_context(command):
                    group.invoke(ctx)

        # Generate summary
        results: Counter[CommandResult] = Counter()
        for execution in self.executions:
            results[execution.result] += 1

        # Log summary
        total = sum(results.values())
        logger.info(
            "Bulk execution complete. Total: %d, Succeeded: %d, Failed: %d, Skipped: %d",
            total,
            results[CommandResult.SUCCESS],
            results[CommandResult.FAILURE],
            results[CommandResult.SKIPPED],
        )

        # In CONTINUE mode, raise error if any commands failed
        if self.mode == BulkRunnerMode.CONTINUE and results[CommandResult.FAILURE] > 0:
            failed_commands = [
                f"Line {e.line_number}: [command]{e.command}[/] [i]({e.error})[/]"
                for e in self.executions
                if e.result == CommandResult.FAILURE
            ]
            raise CommandFileError(
                f"{results[CommandResult.FAILURE]} commands failed:\n"
                + "\n".join(failed_commands)
            )

        return results

    def load_command_file(self) -> list[BulkCommand]:
        """Parse the contents of a command file."""
        try:
            contents = read_file(self.file)
        except Exception as e:
            raise CommandFileError(f"Could not read command file: {e}") from e

        commands: list[BulkCommand] = []

        def add_skipped(
            line: str, line_number: int, error: Optional[BaseException] = None
        ) -> None:
            self.skipped.append(
                CommandExecution(
                    BulkCommand(line=line, line_number=line_number),
                    CommandResult.SKIPPED,
                    error=error,
                    line_number=line_number,
                )
            )

        for lineno, line in enumerate(contents.splitlines(), start=1):
            try:
                command = BulkCommand.from_line(line, line_number=lineno)
                commands.append(command)
            except SkippableLine:
                logger.debug("Skipping line %d: %s", lineno, line)
                add_skipped(line, lineno)
            except Exception as e:
                if self.mode == BulkRunnerMode.SKIP:
                    add_skipped(line, lineno, e)
                    warning(
                        f"Ignoring invalid line {lineno}: [i default]{line}[/] ({e})"
                    )
                else:
                    raise CommandFileError(
                        f"Unable to parse line {lineno} '{line}': {e}"
                    ) from e

        return commands


def run_bulk(ctx: typer.Context, file: Path, mode: BulkRunnerMode) -> None:
    state = get_state()
    runner = BulkRunner(ctx, file, mode)
    try:
        state.bulk = True
        runner.run_bulk()
    finally:
        state.bulk = False
        state.logout_on_exit()
        logger.debug("Bulk execution complete.")



"""Commands to view and manage macros."""

from __future__ import annotations

from typing import Optional

import typer

from zabbix_cli.app import Example
from zabbix_cli.app import app
from zabbix_cli.commands.common.args import get_limit_option
from zabbix_cli.config.constants import OutputFormat
from zabbix_cli.exceptions import ZabbixCLIError
from zabbix_cli.exceptions import ZabbixNotFoundError
from zabbix_cli.output.console import exit_err
from zabbix_cli.output.render import render_result

HELP_PANEL_HOST = "Macro (Host)"
HELP_PANEL_GLOBAL = "Macro (Global)"


def fmt_macro_name(macro: str) -> str:
    """Format macro name for use in a query."""
    macro = macro.strip()
    if not macro:
        # TODO: More specific exception class
        raise ZabbixCLIError("Macro name cannot be empty.")
    if not macro.isupper():
        macro = macro.upper()
    if not macro.startswith("{"):
        macro = "{" + macro
    if not macro.endswith("}"):
        macro = macro + "}"
    if not macro[1] == "$":  # NOTE: refactor could break this
        macro = "{$" + macro[1:]
    if macro == "{$}":
        raise ZabbixCLIError(f"Invalid macro name {macro!r}")
    return macro


@app.command("define_global_macro", rich_help_panel=HELP_PANEL_GLOBAL)
def define_global_macro(
    ctx: typer.Context,
    name: str = typer.Argument(help="Name of the macro", show_default=False),
    value: str = typer.Argument(help="Value of the macro", show_default=False),
) -> None:
    """Create a global macro."""
    from zabbix_cli.commands.results.macro import GlobalMacroResult
    from zabbix_cli.models import Result

    name = fmt_macro_name(name)

    try:
        macro = app.state.client.get_global_macro(macro_name=name)
    except ZabbixNotFoundError:
        pass
    else:
        exit_err(f"Macro {name!r} already exists with value {macro.value!r}")

    macro_id = app.state.client.create_global_macro(macro=name, value=value)
    render_result(
        Result(
            message=f"Created macro {name!r} with ID {macro_id}.",
            result=GlobalMacroResult(macro=name, globalmacroid=macro_id, value=value),
        ),
    )


@app.command("show_global_macros", rich_help_panel=HELP_PANEL_GLOBAL)
def show_global_macros(ctx: typer.Context) -> None:
    """Show all global macros."""
    from zabbix_cli.commands.results.macro import GlobalMacroResult
    from zabbix_cli.models import AggregateResult

    macros = app.state.client.get_global_macros()
    render_result(
        AggregateResult(
            result=[
                GlobalMacroResult(
                    macro=m.macro, globalmacroid=m.globalmacroid, value=m.value
                )
                for m in macros
            ]
        )
    )


@app.command(
    "define_host_macro",
    rich_help_panel=HELP_PANEL_HOST,
    examples=[
        Example(
            "Create a macro named {$SNMP_COMMUNITY} for a host",
            "define_host_macro foo.example.com '{$SNMP_COMMUNITY}' public",
        ),
        Example(
            "Create a macro named {$SITE_URL} for a host (automatic name conversion)",
            "define_host_macro foo.example.com site_url https://example.com",
        ),
    ],
    help="Define a host macro.",
)
@app.command(
    # old name for backwards compatibility
    "define_host_usermacro",
    hidden=True,
    deprecated=True,
    rich_help_panel=HELP_PANEL_HOST,
    help="DEPRECATED: Use add_template_to_group instead.",
)
def define_host_macro(
    # NOTE: should this use old style args?
    hostname: str = typer.Argument(
        help="Host to define macro for.", show_default=False
    ),
    macro_name: str = typer.Argument(
        help=(
            "Name of macro. "
            "Names will be converted to the Zabbix format, "
            "i.e. [value]site_url[/] becomes [value]{$SITE_URL}[/]."
        ),
        show_default=False,
    ),
    macro_value: str = typer.Argument(
        help="Default value of macro.", show_default=False
    ),
) -> None:
    """Create or update a host usermacro."""
    from zabbix_cli.models import Result

    host = app.state.client.get_host(hostname)
    macro_name = fmt_macro_name(macro_name)

    # Determine if we should create or update macro
    try:
        macro = app.state.client.get_macro(host=host, macro_name=macro_name)
    except ZabbixNotFoundError:
        macro_id = app.state.client.create_host_macro(
            host=host, macro=macro_name, value=macro_value
        )
        action = "Created"
    else:
        macro_id = app.state.client.update_macro(
            macroid=macro.hostmacroid, value=macro_value
        )
        action = "Updated"

    render_result(
        Result(
            message=f"{action} macro {macro_name!r} with ID {macro_id} for host {hostname!r}."
        )
    )


@app.command(
    name="show_host_macros",
    rich_help_panel=HELP_PANEL_HOST,
    help="Show all macros defined for a host.",
)
@app.command(
    name="show_host_usermacros",
    rich_help_panel=HELP_PANEL_HOST,
    hidden=True,
    deprecated=True,
    help="DEPRECATED: Use show_host_macros instead.",
)
def show_host_macros(
    hostname_or_id: str = typer.Argument(
        help="Hostname or ID to show macros for",
        show_default=False,
    ),
) -> None:
    """Show all macros defined for a host."""
    from zabbix_cli.commands.results.macro import ShowHostUserMacrosResult
    from zabbix_cli.models import AggregateResult

    # By getting the macros via the host, we also ensure the host exists.
    host = app.state.client.get_host(hostname_or_id, select_macros=True)

    render_result(
        AggregateResult(
            result=[
                ShowHostUserMacrosResult.from_result(macro)
                # Sort macros by name when rendering
                for macro in sorted(host.macros, key=lambda m: m.macro)
            ]
        )
    )


@app.command(
    name="show_macro_hosts",
    rich_help_panel=HELP_PANEL_HOST,
)
@app.command(
    name="show_usermacro_host_list",
    rich_help_panel=HELP_PANEL_HOST,
    hidden=True,
    deprecated=True,
    help="DEPRECATED: Use show_macro_hosts instead.",
)
def show_macro_hosts(
    usermacro: str = typer.Argument(
        help=(
            "Name of macro to find hosts with. "
            "Macro names are automatically formatted, e.g. [value]site_url[/] becomes [value]{$SITE_URL}[/]."
        ),
        show_default=False,
    ),
    limit: Optional[int] = get_limit_option(),
) -> None:
    """Find all hosts with a user macro of the given name.

    Renders a list of the complete macro object and its hosts in JSON mode.
    """
    from zabbix_cli.commands.results.macro import MacroHostListV2
    from zabbix_cli.commands.results.macro import MacroHostListV3
    from zabbix_cli.models import AggregateResult

    usermacro = fmt_macro_name(usermacro)
    macros = app.state.client.get_macros(
        macro_name=usermacro, select_hosts=True, limit=limit
    )
    macros = [macro for macro in macros if macro.hosts]

    # This is a place where we need to differentiate between legacy and
    # new JSON modes instead of sharing a single model and
    # letting the render function figure it out.
    # The V2 command only renders a single host, but the whole point of this
    # command is to list _all_ hosts with the given macro, so we want to render
    # the macro and a list of EVERY host with that macro.
    if (
        app.state.config.app.output.format == OutputFormat.JSON
        and app.state.config.app.legacy_json_format
    ):
        render_result(
            AggregateResult(result=[MacroHostListV2(macro=macro) for macro in macros])
        )
    else:
        if not macros:
            exit_err(f"Macro {usermacro!r} not found.")
        render_result(
            AggregateResult(result=[MacroHostListV3(macro=macro) for macro in macros])
        )


@app.command(
    "define_template_macro",
    rich_help_panel=HELP_PANEL_HOST,
    examples=[
        Example(
            "Create a macro named {$SNMP_COMMUNITY} with the value 'public' for a template",
            "define_template_macro Mytemplate '{$SNMP_COMMUNITY}' public",
        ),
        Example(
            "Create a macro named {$SITE_URL} with an URL value for a template (automatic name conversion)",
            "define_template_macro Mytemplate site_url https://example.com",
        ),
    ],
    help="Define a template macro.",
)
def define_template_macro(
    template_name: str = typer.Argument(
        help="Name of template to define macro for.", show_default=False
    ),
    macro_name: str = typer.Argument(
        help=(
            "Name of macro. "
            "Names will be converted to the Zabbix format, "
            "i.e. [value]site_url[/] becomes [value]{$SITE_URL}[/]."
        ),
        show_default=False,
    ),
    macro_value: str = typer.Argument(
        help="Default value of macro.", show_default=False
    ),
) -> None:
    """Create or update a template macro."""
    from zabbix_cli.models import Result

    template = app.state.client.get_template(template_name)
    macro_name = fmt_macro_name(macro_name)

    # Determine if we should create or update macro
    try:
        macro = app.state.client.get_macro(template=template, macro_name=macro_name)
    except ZabbixNotFoundError:
        macro_id = app.state.client.create_template_macro(
            template=template, macro=macro_name, value=macro_value
        )
        action = "Created"
    else:
        macro_id = app.state.client.update_macro(
            macroid=macro.hostmacroid, value=macro_value
        )
        action = "Updated"

    render_result(
        Result(
            message=f"{action} macro {macro_name!r} with ID {macro_id} for template {template}."
        )
    )


@app.command(
    name="show_template_macros",
    rich_help_panel=HELP_PANEL_HOST,
    help="Show all macros defined for a template.",
)
def show_template_macros(
    template_name_or_id: str = typer.Argument(
        help="Template name or ID to show macros for",
        show_default=False,
    ),
) -> None:
    """Show all macros defined for a host."""
    from zabbix_cli.commands.results.macro import ShowHostUserMacrosResult
    from zabbix_cli.models import AggregateResult

    # By getting the macros via the template, we also ensure the template exists.
    template = app.state.client.get_template(template_name_or_id, select_macros=True)

    render_result(
        AggregateResult(
            result=[
                ShowHostUserMacrosResult.from_result(macro)
                # Sort macros by name when rendering
                for macro in sorted(template.macros, key=lambda m: m.macro)
            ]
        )
    )


@app.command(
    "show_macro_templates",
    rich_help_panel=HELP_PANEL_HOST,
    examples=[
        Example(
            "Show all templates with a user macro named {$SNMP_COMMUNITY}",
            "show_macro_templates SNMP_COMMUNITY",
        )
    ],
)
@app.command(
    "show_usermacro_template_list",
    rich_help_panel=HELP_PANEL_HOST,
    hidden=True,
    deprecated=True,
    help="DEPRECATED: Use show_macro_templates instead.",
)
def show_macro_templates(
    ctx: typer.Context,
    macro_name: str = typer.Argument(
        help="Name of the macro to find templates with. Automatically formatted.",
        show_default=False,
    ),
    limit: Optional[int] = get_limit_option(),
) -> None:
    """Find all templates with a macro of the given name."""
    import itertools

    from zabbix_cli.commands.results.macro import ShowUsermacroTemplateListResult
    from zabbix_cli.models import AggregateResult

    macro_name = fmt_macro_name(macro_name)
    macros = app.state.client.get_macros(
        macro_name=macro_name, select_templates=True, limit=limit
    )
    macros = [macro for macro in macros if macro.templates]

    results = itertools.chain.from_iterable(
        [
            [
                ShowUsermacroTemplateListResult(
                    macro=macro.macro,
                    value=macro.value,
                    templateid=template.templateid,
                    template=template.host,
                )
                for template in macro.templates
            ]
            for macro in macros
        ]
    )

    render_result(
        AggregateResult(
            result=list(results),
        )
    )



from __future__ import annotations

from datetime import datetime
from datetime import timedelta

import pytest
from freezegun import freeze_time
from zabbix_cli.utils import convert_duration
from zabbix_cli.utils.utils import convert_time_to_interval
from zabbix_cli.utils.utils import convert_timestamp
from zabbix_cli.utils.utils import convert_timestamp_interval


@pytest.mark.parametrize(
    "input,expect",
    [
        (
            "1 hour",
            timedelta(hours=1),
        ),
        (
            "1 hour 30 minutes",
            timedelta(hours=1, minutes=30),
        ),
        (
            "1 hour 30 minutes 30 seconds",
            timedelta(hours=1, minutes=30, seconds=30),
        ),
        (
            "1 day 1 hour 30 minutes 30 seconds",
            timedelta(days=1, hours=1, minutes=30, seconds=30),
        ),
        (
            "2 hours 30 seconds",
            timedelta(hours=2, seconds=30),
        ),
        (
            "2h30s",
            timedelta(hours=2, seconds=30),
        ),
        (
            "2 hours 30 minutes 30 seconds",
            timedelta(hours=2, minutes=30, seconds=30),
        ),
        (
            "2 hour 30 minute 30 second",
            timedelta(hours=2, minutes=30, seconds=30),
        ),
        (
            "2h30m30s",
            timedelta(hours=2, minutes=30, seconds=30),
        ),
        (
            "9030s",
            timedelta(seconds=9030),
        ),
        (
            "9030",
            timedelta(seconds=9030),
        ),
        (
            "0",
            timedelta(seconds=0),
        ),
        (
            "0d0h0m0s",
            timedelta(days=0, hours=0, minutes=0, seconds=0),
        ),
    ],
)
def test_convert_duration(input: str, expect: timedelta) -> None:
    assert convert_duration(input) == expect
    assert convert_duration(input).total_seconds() == expect.total_seconds()


@pytest.mark.parametrize(
    "input, expect",
    [
        pytest.param(
            "2016-11-21T22:00 to 2016-11-21T23:00",
            (datetime(2016, 11, 21, 22, 0, 0), datetime(2016, 11, 21, 23, 0, 0)),
            id="Legacy format",  #  (ISO w/o timezone and seconds)
        ),
        pytest.param(
            "2016-11-21T22:00:00 to 2016-11-21T23:00:00",
            (datetime(2016, 11, 21, 22, 0, 0), datetime(2016, 11, 21, 23, 0, 0)),
            id="ISO w/o timezone",
        ),
        pytest.param(
            "2016-11-21 22:00:00 to 2016-11-21 23:00:00",
            (datetime(2016, 11, 21, 22, 0, 0), datetime(2016, 11, 21, 23, 0, 0)),
            id="ISO w/o timezone (space-separated)",
        ),
    ],
)
def test_convert_timestamp_interval(
    input: str, expect: tuple[datetime, datetime]
) -> None:
    assert convert_timestamp_interval(input) == expect
    # TODO: test with mix of formats. e.g. "2016-11-21T22:00 to 2016-11-21 23:00:00"


@pytest.mark.parametrize(
    "input,expect",
    [
        pytest.param(
            "2016-11-21T22:00",
            datetime(2016, 11, 21, 22, 0, 0),
            id="Legacy",
        ),
        pytest.param(
            "2016-11-21T22:00:00",
            datetime(2016, 11, 21, 22, 0, 0),
            id="ISO w/o timezone",
        ),
        pytest.param(
            "2016-11-21 22:00:00",
            datetime(2016, 11, 21, 22, 0, 0),
            id="ISO w/o timezone (space-separated)",
        ),
    ],
)
def test_convert_timestamp(input: str, expect: datetime) -> None:
    assert convert_timestamp(input) == expect


@pytest.mark.parametrize(
    "input,expect_duration",
    [
        pytest.param(
            "2016-11-21T22:00 to 2016-11-21T23:00",
            timedelta(hours=1),
            id="Range: Legacy",
        ),
        pytest.param(
            "2016-11-21T22:00:00 to 2016-11-21T23:00:00",
            timedelta(hours=1),
            id="Range: ISO w/o timezone",
        ),
        pytest.param(
            "2016-11-21 22:00:00 to 2016-11-21 23:00:00",
            timedelta(hours=1),
            id="Range: ISO w/o timezone (space-separated)",
        ),
        pytest.param(
            "1 hour",
            timedelta(hours=1),
            id="Duration: 1h (long form)",
        ),
        pytest.param(
            "1 day 1 hour 30 minutes 30 seconds",
            timedelta(days=1, hours=1, minutes=30, seconds=30),
            id="Duration: 1d1h30m30s (long form)",
        ),
        pytest.param(
            "1h",
            timedelta(hours=1),
            id="Duration: 1h (short form)",
        ),
        pytest.param(
            "1d1h30m30s",
            timedelta(days=1, hours=1, minutes=30, seconds=30),
            id="Duration: 1d1h30m30s (short form)",
        ),
    ],
)
@freeze_time("2016-11-21 22:00:00")
def test_convert_time_to_interval(input: str, expect_duration: timedelta) -> None:
    start, end = convert_time_to_interval(input)
    assert start == datetime(2016, 11, 21, 22, 0, 0)
    assert end == start + expect_duration



from __future__ import annotations

import logging
import subprocess
import sys
from collections.abc import Iterable
from collections.abc import Iterator
from collections.abc import Sequence
from datetime import datetime
from enum import IntEnum
from enum import auto
from functools import wraps
from pathlib import Path
from subprocess import CompletedProcess
from typing import Any
from typing import NamedTuple
from typing import NoReturn
from typing import Optional
from typing import Protocol

import typer
from rich.console import Console
from rich.panel import Panel
from strenum import StrEnum

console = Console()
err_console = Console(stderr=True, style="red")


class ExitCode(IntEnum):
    """Exit codes for the script."""

    ERROR = 1
    UNHANDLED_EXCEPTION = 3
    INTERRUPTED = 130


class VersionType(StrEnum):
    major = "major"
    minor = "minor"
    patch = "patch"


class ReleaseStage(StrEnum):
    # Release
    release = "release"

    # Alpha
    a = "a"
    alpha = "alpha"

    # Beta
    b = "b"
    beta = "beta"

    # Release Candidate
    c = "c"
    rc = "rc"
    pre = "pre"
    preview = "preview"

    # Revision / Post
    r = "r"
    rev = "rev"
    post = "post"

    # Dev
    dev = "dev"


class CommandCheck(NamedTuple):
    program: str
    command: Sequence[str]
    message: str


REQUIRED_COMMANDS = [
    CommandCheck(
        program="Hatch",
        command=["hatch", "--version"],
        message="Hatch is not installed. Please install it with `pip install hatch`.",
    ),
    CommandCheck(
        program="Git",
        command=["git", "--version"],
        message="Git is not installed. Please install it.",
    ),
]


def check_commands() -> None:
    """Checks that we have the necessary programs installed and available"""
    for command in REQUIRED_COMMANDS:
        try:
            subprocess.check_output(command.command)
        except FileNotFoundError:
            exit_err(f"{command.message} :x:", style="bold red")
        else:
            console.print(f"{command.program} :white_check_mark:")


def exit_err(
    msg: str, code: int = ExitCode.ERROR, style: Optional[str] = None
) -> NoReturn:
    """Print an error message and exit with the given code."""
    err_console.print(msg, style=style)
    sys.exit(code)


class State(IntEnum):
    OLD_VERSION = auto()  # before bumping version
    NEW_VERSION = auto()  # after bumping version
    MODIFY_CHANGELOG = auto()  # after modifying changelog
    GIT_ADD = auto()  # after git adding version file
    GIT_COMMIT = auto()  # after git commit
    GIT_TAG = auto()  # after git tag
    GIT_PUSH = auto()  # after git push

    @classmethod
    def __missing__(cls, value: Any) -> State:
        return State.OLD_VERSION

    @classmethod
    def ensure_contiguous(cls) -> None:
        """Ensure that enum values are contiguous and increment by 1"""
        # NOTE: this is just a sanity check that ensures we increment and
        # decrement the state machine correctly.
        if len(cls) != max(cls):  # starts at 1
            raise ValueError("Enum values must be contiguous")
        prev = None
        for st in cls:
            if prev is not None and st != prev + 1:
                raise ValueError("Enum values must increment by 1")
            prev = st


class StateMachine:
    def __init__(self) -> None:
        self.state = State.OLD_VERSION

    def forward(self) -> None:
        """Advance the state machine to the next state."""
        self.state = State(self.state.value + 1)

    def back(self) -> State:
        """Revert the state machine to the previous state."""
        self.state = State(self.state.value - 1)
        return self.state

    def rewind(self) -> Iterator[State]:
        """Iterate through all states in reverse order from the current state."""
        yield self.state
        while self.state != State.OLD_VERSION:
            yield self.back()


# I will not add proper type annotations to this decorator. If someone wants to do it, go ahead!
def advance(after: State) -> Any:
    """Advance state machine after a function call and check the expected state."""

    def decorator(f: Any) -> Any:
        @wraps(f)
        def inner(self: StateMachine, *args: Any, **kwargs: Any) -> Any:
            res = f(self, *args, **kwargs)
            # Only advance if we haven't run the function before.
            # In certain cases, we have to run git add twice, in which case
            # we cannot advance and and check the state on the second call.
            if after > self.state:
                self.forward()
                assert self.state == after, f"Expected state {after}, got {self.state}"
            return res

        return inner

    return decorator


class Runner(Protocol):
    def __call__(
        self, args: Sequence[str], *aargs: Any, **kwargs: Any
    ) -> CompletedProcess[bytes]: ...


class VersionBumper(StateMachine):
    run: Runner
    """Run a command in a subprocess.

    If dry_run is True, the command will be printed but not executed."""

    source_dir = Path("zabbix_cli")
    version_file = source_dir / "__about__.py"
    changelog_file = Path("CHANGELOG")
    changelog_file_bak = changelog_file.with_suffix(".bak")
    old_version: str | None = None
    new_version: str | None = None

    def __init__(
        self, target_version: str, push: bool, dry_run: bool, remote: str, branch: str
    ) -> None:
        super().__init__()
        self.target_version = target_version
        self.dry_run = dry_run
        self.push = push and not dry_run
        self.remote = remote
        self.branch = branch
        self.run = self.get_runner(dry_run)

    @property
    def tag(self) -> str:
        return f"{self.new_version}"

    def get_runner(self, dry_run: bool) -> Runner:
        def dryrun_subprocess_run(
            args: Sequence[str], *aargs: Iterable[str], **kwargs: str
        ) -> CompletedProcess[bytes]:
            """Wrapper around subprocess.run that prints the command and returns a dummy CompletedProcess"""
            args_quoted = [f"'{a}'" if " " in a else a for a in args]
            print(f"Running: {' '.join(args_quoted)}")
            return subprocess.CompletedProcess(
                args=args, returncode=0, stdout=b"", stderr=b""
            )

        if not dry_run:
            return subprocess.run

        lines = [
            "[bold]Running in dry-run mode.[/bold]",
            "Commands will not be executed.",
            "",
            "[bold]NOTE:[/bold] The previewed commands will show the current package version!",
        ]

        warning_console = Console(stderr=True, style="yellow")
        warning_console.print(
            Panel("\n".join(lines), title="Dry-run mode", expand=False)
        )
        return dryrun_subprocess_run

    def undo(self) -> None:
        for st in self.rewind():
            # from last to first
            # Best-effort cleanup
            try:
                if st == State.GIT_PUSH:
                    # probably nothing to clean up here
                    # we could do a git push --delete, but if it failed,
                    # it probably isn't in the upstream repo anyway
                    pass
                elif st == State.GIT_TAG:
                    if not self.new_version:
                        raise ValueError("No new version to untag.")
                    self.run(["git", "tag", "-d", self.tag])
                elif st == State.GIT_COMMIT:
                    # Undo last commit, but keep changes in the working directory
                    # --soft to keep changes in the working directory
                    self.run(["git", "reset", "--soft", "HEAD~"])
                elif st == State.GIT_ADD:
                    # Unstage changelog
                    self.run(["git", "restore", "--staged", str(self.changelog_file)])
                elif st == State.MODIFY_CHANGELOG:
                    # Revert the changes in the changelog file
                    if self.changelog_file_bak.exists():
                        self.changelog_file_bak.rename(self.changelog_file)
                elif st == State.NEW_VERSION:
                    # Hatch can't set the version to a lower than the current version
                    # so we have to revert the changes in the version file
                    self.run(["git", "checkout", "HEAD", str(self.version_file)])
            except Exception as e:
                print(f"Failed to revert state {self.state}: {e}", file=sys.stderr)
                raise e

    def cleanup(self) -> None:
        if self.changelog_file_bak.exists():
            self.changelog_file_bak.unlink()

    def bump(self) -> None:
        check_commands()
        self.state.ensure_contiguous()  # sanity check

        # TODO: * ensure we are on the correct branch
        #           `git rev-parse --abbrev-ref HEAD`
        #       * ensure we are up-to-date with the remote
        #           `git fetch upstream`

        # TODO: add decorator that advances state and checks expected state

        old_version = self.get_project_version()
        self.old_version = old_version

        self.set_version(self.target_version)
        # We have to re-fetch the new version after bumping, because
        # we might have used a version strategy rather than a specific version.
        # i.e. "major", "minor", "patch", "rc", etc.
        self.new_version = self.get_project_version()
        self.add_changelog_header(self.new_version)
        self.git_add()
        self.git_commit(self.new_version)
        self.git_tag(self.new_version)
        if self.push:
            self.git_push()

    def get_project_version(self) -> str:
        """Get the current project version from Hatch."""
        # NOTE: This is run directly with subprocess.check_output
        # instead of going through the run method, because we want to
        # actually call this command in dry-run mode as well.
        try:
            new_version = subprocess.check_output(["hatch", "version"]).decode().strip()
        except Exception as e:
            exit_err(f"Failed to get project version from Hatch: {e}")
        return new_version

    @advance(State.GIT_ADD)
    def git_add(self, *files: str) -> CompletedProcess[bytes]:
        p_git_add = self.run(
            [
                "git",
                "add",
                str(self.version_file.absolute()),
                str(self.changelog_file.absolute()),
                *files,
            ],
            capture_output=True,
        )
        if p_git_add.returncode != 0:
            exit_err(f"Failed to stage files: {p_git_add.stderr.decode()}")
        return p_git_add

    def get_pre_commit_files(self, msg: str) -> list[str]:
        files: list[str] = []
        for line in msg.splitlines():
            if line.startswith("Fixing "):
                _, _, file = line.partition("Fixing ")
                files.append(file)
        return files

    @advance(State.GIT_COMMIT)
    def git_commit(
        self, new_version: str, *, rerun: bool = False
    ) -> CompletedProcess[bytes]:
        p_git_commit = self.run(
            ["git", "commit", "-m", f"Bump version to {new_version}"],
            capture_output=True,
        )
        if p_git_commit.returncode != 0:
            # pre-commit might have modified our files
            msg = p_git_commit.stderr.decode()
            if "- hook id" in msg and not rerun:
                # re-run git-add and git-commit
                self.state = State(State.GIT_ADD - 1)
                # User might have added other files to the staging area
                # that pre-commit modified. We need to detect these files
                # and add them to the commit.
                files = self.get_pre_commit_files(msg)
                self.git_add(*files)
                self.git_commit(new_version, rerun=True)
            else:
                exit_err(f"Failed to commit version bump:\n{msg}")
        return p_git_commit

    @advance(State.GIT_TAG)
    def git_tag(self, new_version: str) -> CompletedProcess[bytes]:
        p_git_tag = self.run(["git", "tag", self.tag], capture_output=True)
        if p_git_tag.returncode != 0:
            exit_err(f"Failed to tag version: {p_git_tag.stderr.decode()}")
        return p_git_tag

    @advance(State.GIT_PUSH)
    def git_push(self) -> CompletedProcess[bytes]:
        p_git_push = self.run(
            ["git", "push", "--tags", self.remote, self.branch], capture_output=True
        )
        if p_git_push.returncode != 0:
            exit_err(f"Failed to push new version: {p_git_push.stderr.decode()}")
        return p_git_push

    @advance(State.MODIFY_CHANGELOG)
    def add_changelog_header(self, new_version: str) -> None:
        if self.dry_run:
            console.print(
                f"Would add changelog header for version {new_version} to {self.changelog_file}"
            )
            return

        changelog = self.changelog_file.read_text()
        self.changelog_file_bak.write_text(changelog)

        # Find the line containing the unreleased header
        lines = changelog.splitlines()
        index = next(
            iter(
                [
                    idx
                    for idx, line in enumerate(lines)
                    if line.startswith("## Unreleased")
                ],
            ),
            None,
        )
        if index is None:
            exit_err(f"Failed to find '## Unreleased' section in {self.changelog_file}")

        header = f"## [{new_version}](https://github.com/unioslo/zabbix-cli/tree/{new_version}) - {datetime.now().strftime('%Y-%m-%d')}"
        lines[index] = "<!-- ## Unreleased -->"  # comment out
        lines.insert(index + 1, f"\n{header}")  # insert after
        self.changelog_file.write_text("\n".join(lines))

    @advance(State.NEW_VERSION)
    def set_version(self, version: str) -> str:
        # We don't verify that the version arg is valid, we just pass it
        # to hatch and let it handle it.
        # Worst case scenario, we get a non-zero exit code and the script exits
        p_version = self.run(["hatch", "version", version], capture_output=True)
        if p_version.returncode != 0:
            exit_err(f"Failed to set version: {p_version.stderr.decode()}")
        return p_version.stdout.decode().strip()  # the new version


def main(
    version: str = typer.Argument(
        ...,
        help="Version bump to perform or new version to set.",
        metavar="["
        + "|".join(VersionType)
        + "|x.y.z],["
        + "|".join(ReleaseStage)
        + "]",
        show_default=False,
    ),
    push: bool = typer.Option(
        True,
        "--push/--no-push",
        help="Push the created tag and commmit to the remote repository automatically.",
    ),
    dry_run: bool = typer.Option(
        False,
        "--dry-run",
        help="Print the commands that would be run without executing them.",
        is_flag=True,
    ),
    branch: str = typer.Option(
        "master",
        "--branch",
        help="The remote branch to push to.",
    ),
    remote: str = typer.Option(
        "upstream",
        "--remote",
        help="The remote repository to push to.",
    ),
) -> None:
    """Bump the version of the project and create a new git tag.

    If using --no-push, remember to also push the tag manually:
    `git push --tags upstream main`.

    Examples:
    $ python bump_version.py minor

    $ python bump_version.py major,rc

    $ python bump_version.py 1.2.3 # generally don't use this
    """
    bumper = VersionBumper(
        target_version=version,
        push=push,
        dry_run=dry_run,
        remote=remote,
        branch=branch,
    )
    try:
        bumper.bump()
    except KeyboardInterrupt:
        bumper.undo()
        exit_err("Bump cancelled by user.", code=ExitCode.INTERRUPTED)
    except Exception as e:
        bumper.undo()
        logging.exception(e)
        sys.exit(ExitCode.UNHANDLED_EXCEPTION)
    else:
        console.print(
            f"Successfully bumped version from {bumper.old_version} to {bumper.new_version} :tada:"
        )
    finally:
        bumper.cleanup()


if __name__ == "__main__":
    typer.run(main)



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Optional

from pydantic import BaseModel
from pydantic import Field
from pydantic import model_serializer
from typing_extensions import Self

from zabbix_cli.models import ColsRowsType
from zabbix_cli.models import MetaKey
from zabbix_cli.models import TableRenderable
from zabbix_cli.pyzabbix.types import Host
from zabbix_cli.pyzabbix.types import HostList
from zabbix_cli.pyzabbix.types import Proxy
from zabbix_cli.pyzabbix.types import ProxyGroup

if TYPE_CHECKING:
    from zabbix_cli.models import ColsRowsType
    from zabbix_cli.models import RowsType


class BaseHostProxyResult(TableRenderable):
    source: str = Field(..., json_schema_extra={MetaKey.HEADER: "Source Proxy"})
    """Name of the old proxy."""
    hosts: list[str] = []
    """Name of the hosts that were updated."""


class UpdateHostProxyResult(BaseHostProxyResult):
    """Result type for `update_host_proxy` command."""

    destination: str = Field(
        ..., json_schema_extra={MetaKey.HEADER: "Destination Proxy"}
    )
    """Name of the new proxy"""

    @classmethod
    def from_result(
        cls,
        hosts: list[Host],
        source_proxy: Optional[Proxy],
        dest_proxy: Optional[Proxy],
    ) -> Self:
        return cls(
            source=source_proxy.name if source_proxy else "",
            destination=dest_proxy.name if dest_proxy else "",
            hosts=[h.host for h in hosts],
        )


class ClearHostProxyResult(BaseHostProxyResult):
    """Result type for `update_host_proxy` command."""

    @classmethod
    def from_result(
        cls,
        hosts: list[Host],
        source_proxy: Optional[Proxy],
    ) -> Self:
        return cls(
            source=source_proxy.name if source_proxy else "",
            hosts=[h.host for h in hosts],
        )


class MoveProxyHostsResult(TableRenderable):
    """Result type for `move_proxy_hosts` command."""

    source: Optional[str] = None
    """ID of the source (old) proxy."""
    destination: Optional[str] = None
    """ID of the destination (new) proxy."""
    hosts: list[str] = []


class LBProxy(BaseModel):
    """A load balanced proxy."""

    proxy: Proxy
    hosts: list[Host] = []
    weight: int
    count: int = 0

    @model_serializer
    def ser_model(self):
        return {
            "name": self.proxy.name,
            "proxyid": self.proxy.proxyid,
            "weight": self.weight,
            "count": self.count,
            "hosts": [h.host for h in self.hosts],
        }


class LBProxyResult(TableRenderable):
    """Result type for `load_balance_proxy_hosts` command."""

    proxies: list[LBProxy]

    def __cols_rows__(self) -> ColsRowsType:
        cols = ["Proxy", "Weight", "Hosts"]
        rows: RowsType = []
        for proxy in self.proxies:
            rows.append([proxy.proxy.name, str(proxy.weight), str(len(proxy.hosts))])
        return cols, rows


class UpdateHostGroupProxyResult(TableRenderable):
    """Result type for `update_hostgroup_proxy` command."""

    proxy: str
    hosts: list[str] = []

    @classmethod
    def from_result(cls, proxy: Proxy, hosts: list[Host]) -> Self:
        return cls(proxy=proxy.name, hosts=[host.host for host in hosts])


class UpdateHostGroupProxyGroupResult(TableRenderable):
    """Result type for `update_hostgroup_proxygroup` command."""

    proxy_group: str
    hosts: list[str] = []

    @classmethod
    def from_result(cls, proxy_group: ProxyGroup, hosts: list[Host]) -> Self:
        return cls(proxy_group=proxy_group.name, hosts=[host.host for host in hosts])


class ShowProxiesResult(TableRenderable):
    """Result type for `show_proxy` command."""

    proxy: Proxy
    show_hosts: bool = Field(default=False, exclude=True)

    @classmethod
    def from_result(cls, proxy: Proxy, *, show_hosts: bool = False) -> Self:
        return cls(proxy=proxy, show_hosts=show_hosts)

    @property
    def hosts_fmt(self) -> str:
        if self.show_hosts:
            return ", ".join(f"{host.host}" for host in self.proxy.hosts)
        else:
            return str(len(self.proxy.hosts))

    def __cols_rows__(self) -> ColsRowsType:
        cols = [
            "Name",
            "Address",
            "Mode",
            "Hosts",
        ]

        rows: RowsType = [
            [
                self.proxy.name,
                str(self.proxy.address),
                self.proxy.mode,
                self.hosts_fmt,
            ]
        ]
        if self.zabbix_version.release >= (7, 0, 0):
            cols.extend(["Version", "Compatibility"])
            rows[0].extend([str(self.proxy.version), self.proxy.compatibility_rich])
        return cols, rows


class ShowProxyHostsResult(TableRenderable):
    proxy: Proxy
    hosts: HostList = []

    @classmethod
    def from_result(cls, proxy: Proxy) -> Self:
        # HACK: remove the list of hosts from the proxy and store it separately
        # so that we can render the proxy separately from the hosts
        hosts = proxy.hosts
        proxy.hosts = []
        return cls(proxy=proxy, hosts=hosts)

    def __cols_rows__(self) -> ColsRowsType:
        cols = ["Proxy", "Hosts"]
        rows: RowsType = [
            [
                self.proxy.name,
                "\n".join(h.host for h in self.hosts),
            ]
        ]
        return cols, rows


class ShowProxyGroupHostsResult(TableRenderable):
    proxy_group: ProxyGroup
    hosts: HostList = []

    def __cols_rows__(self) -> ColsRowsType:
        cols = ["Proxy Group", "Hosts"]
        rows: RowsType = [
            [
                self.proxy_group.name,
                "\n".join(h.host for h in self.hosts),
            ]
        ]
        return cols, rows



from __future__ import annotations

import random
import re
from typing import TYPE_CHECKING
from typing import Optional

from zabbix_cli.exceptions import ZabbixAPICallError
from zabbix_cli.exceptions import ZabbixNotFoundError
from zabbix_cli.pyzabbix.client import ZabbixAPI

if TYPE_CHECKING:
    from zabbix_cli.pyzabbix.types import Proxy


def get_random_proxy(client: ZabbixAPI, pattern: Optional[str] = None) -> Proxy:
    """Fetch a random proxy, optionally matching a regex pattern."""
    proxies = client.get_proxies()
    if not proxies:
        raise ZabbixNotFoundError("No proxies found")
    if pattern:
        try:
            re_pattern = re.compile(pattern)
        except re.error as e:
            raise ZabbixAPICallError(f"Invalid proxy regex pattern: {pattern!r}") from e
        proxies = [proxy for proxy in proxies if re_pattern.match(proxy.name)]
        if not proxies:
            raise ZabbixNotFoundError(f"No proxies matching pattern {pattern!r}")
    return random.choice(proxies)


def get_proxy_map(client: ZabbixAPI) -> dict[str, Proxy]:
    """Fetch all proxies and return a mapping of proxy IDs to Proxy objects."""
    proxies = client.get_proxies()
    return {proxy.proxyid: proxy for proxy in proxies}
